# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=protected-access

from functools import singledispatch
from pathlib import Path
from typing import Any

from azure.ai.ml._restclient.v2023_08_01_preview.models import JobBase as JobBaseData
from azure.ai.ml._restclient.v2025_01_01_preview.models import JobBase as JobBaseData202501
from azure.ai.ml.constants._common import DEFAULT_EXPERIMENT_NAME
from azure.ai.ml.entities._builders.command import Command
from azure.ai.ml.entities._builders.pipeline import Pipeline
from azure.ai.ml.entities._builders.spark import Spark
from azure.ai.ml.entities._builders.sweep import Sweep
from azure.ai.ml.entities._job.job_name_generator import generate_job_name

from .import_job import ImportJob
from .job import Job


def generate_defaults(job: Job, rest_job: JobBaseData) -> None:
    # Default name to a generated user friendly name.
    if not job.name:
        rest_job.name = generate_job_name()

    if not job.display_name:
        rest_job.properties.display_name = rest_job.name

    # Default experiment to current folder name or "Default"
    if not job.experiment_name:
        rest_job.properties.experiment_name = Path("./").resolve().stem.replace(" ", "") or DEFAULT_EXPERIMENT_NAME


@singledispatch
def to_rest_job_object(something: Any) -> JobBaseData:
    raise NotImplementedError()


@to_rest_job_object.register(Job)
def _(job: Job) -> JobBaseData:
    # TODO: Bug Item number: 2883432
    rest_job = job._to_rest_object()  # type: ignore
    generate_defaults(job, rest_job)
    return rest_job


@to_rest_job_object.register(Command)
def _(command: Command) -> JobBaseData202501:
    rest_job = command._to_job()._to_rest_object()
    generate_defaults(command, rest_job)
    return rest_job


@to_rest_job_object.register(Sweep)
def _(sweep: Sweep) -> JobBaseData:
    rest_job = sweep._to_job()._to_rest_object()
    generate_defaults(sweep, rest_job)
    return rest_job


@to_rest_job_object.register(Pipeline)
def _(pipeline: Pipeline) -> JobBaseData:
    rest_job = pipeline._to_job()._to_rest_object()
    generate_defaults(pipeline, rest_job)
    return rest_job


@to_rest_job_object.register(Spark)
def _(spark: Spark) -> JobBaseData:
    rest_job = spark._to_job()._to_rest_object()
    generate_defaults(spark, rest_job)
    return rest_job


@to_rest_job_object.register(ImportJob)
def _(importJob: ImportJob) -> JobBaseData:
    rest_job = importJob._to_rest_object()
    generate_defaults(importJob, rest_job)
    return rest_job



import pytest

from azure.ai.ml import load_workspace
from azure.ai.ml.entities import Hub


@pytest.mark.unittest
@pytest.mark.core_sdk_test
class TestProjectEntity:
    def test_project_schema_manipulation(self) -> None:
        hub = load_workspace(source="./tests/test_configs/workspace/ai_workspaces/aihub_min.yml")
        assert hub is not None
        assert type(hub) == Hub
        assert hub.name == "hub_name"
        assert hub.location == "WestCentralUS"
        assert (
            hub.default_resource_group
            == "/subscriptions/00000000-0000-0000-0000-000000000000/resourceGroups/my-default-resource-group"
        )
        assert hub.tags == {"purpose": "testing", "team": "ws-management"}
        assert (
            hub.storage_account
            == "/subscriptions/00000000-0000-0000-0000-000000000000/resourceGroups/testRG/providers/Microsoft.Storage/storageAccounts/testAccount"
        )
        assert (
            hub.container_registry
            == "/subscriptions/b17253fa-f327-42d6-9686-f3e553e24763/resourcegroups/testRG/providers/Microsoft.ContainerRegistry/registries/testACR"
        )
        assert (
            hub.key_vault
            == "/subscriptions/00000000-0000-0000-0000-000000000000/resourceGroups/testRG/providers/Microsoft.KeyVault/vaults/testKeyVault"
        )
        assert (
            hub.application_insights
            == "/subscriptions/00000000-0000-0000-0000-000000000000/resourceGroups/testRG/providers/Microsoft.Insights/components/testAppIn"
        )
        assert hub.image_build_compute == "some_compute_name"
        assert hub.public_network_access == "Enabled"
        assert hub.hbi_workspace == True
        assert (
            hub.customer_managed_key.key_vault
            == "/subscriptions/00000000-0000-0000-0000-000000000000/resourceGroups/testRG/providers/Microsoft.KeyVault/vaults/testKeyVault2"
        )
        assert hub.customer_managed_key.key_uri == "https://keyvault.vault.azure.net/keys/cmkkey/123456"
        assert (
            hub.serverless_compute.custom_subnet
            == "/subscriptions/00000000-0000-0000-0000-000000000000/resourcegroups/testRG/providers/Microsoft.Network/virtualNetworks/testwsvnet/subnets/testsubnet"
        )
        assert hub.serverless_compute.no_public_ip

    # Note, this test does not ensure as much API correctness as
    # you'd like because of the workspace operation's reliance on ARM
    # template madness.
    @pytest.mark.skip(reason="Rest conversion is/has always been buggy aparently. Not fixing it yet.")
    def test_hub_rest_conversion(self) -> None:
        original_hub = Hub(
            name="hub_name",
            location="somewhere",
            description="test desc",
            tags={"purpose": "testing"},
            display_name="test_display_name",
            resource_group="testRG",
            storage_account="testAccount",
            key_vault="testKeyVault",
            container_registry="testACR",
            image_build_compute="some_compute_name",
            public_network_access="Enabled",
            primary_user_assigned_identity="id1",
            enable_data_isolation=True,
            default_resource_group="my-default-resource-group",
        )

        # Convert hub back and forth from rest equivalent and ensure consistency.
        rest_hub = original_hub._to_rest_object()
        new_hub = Hub._from_rest_object(rest_hub)

        assert new_hub.name == original_hub.name
        assert new_hub.location == original_hub.location
        assert new_hub.description == original_hub.description
        assert new_hub.tags == original_hub.tags
        assert new_hub.display_name == original_hub.display_name
        assert new_hub.resource_group == original_hub.resource_group
        assert new_hub.storage_account == original_hub.storage_account
        assert new_hub.key_vault == original_hub.key_vault
        assert new_hub.container_registry == original_hub.container_registry
        assert new_hub.image_build_compute == original_hub.image_build_compute
        assert new_hub.public_network_access == original_hub.public_network_access
        assert new_hub.primary_user_assigned_identity == original_hub.primary_user_assigned_identity
        assert new_hub.enable_data_isolation == original_hub.enable_data_isolation
        assert new_hub.default_resource_group == original_hub.default_resource_group



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from marshmallow import fields

from azure.ai.ml._schema import StringTransformedEnum
from azure.ai.ml._schema.workspace import WorkspaceSchema
from azure.ai.ml.constants import WorkspaceKind


class ProjectSchema(WorkspaceSchema):
    kind = StringTransformedEnum(required=True, allowed_values=WorkspaceKind.PROJECT)
    hub_id = fields.Str(required=True)



from pathlib import Path

from azure.ai.ml import Input, Output, dsl, load_component, spark
from azure.ai.ml.constants import AssetTypes, InputOutputModes
from azure.ai.ml.entities import PipelineJob

parent_dir = str(Path(__file__).parent)


def generate_dsl_pipeline_from_builder_without_entry() -> PipelineJob:
    # define the spark task
    iris_data = Input(path=parent_dir + "/dataset/iris.csv", type=AssetTypes.URI_FILE, mode=InputOutputModes.DIRECT)
    add_greeting_column_func = spark(
        code=parent_dir + "/src",
        py_files=["utils.zip"],
        files=["my_files.txt"],
        driver_cores=2,
        driver_memory="1g",
        executor_cores=1,
        executor_memory="1g",
        executor_instances=1,
        inputs=dict(file_input=iris_data),
        args="--file_input ${{inputs.file_input}}",
        resources={"instance_type": "standard_e4s_v3", "runtime_version": "3.4.0"},
    )

    # Define pipeline
    @dsl.pipeline(description="submit a pipeline with spark job")
    def spark_pipeline_from_builder(iris_data):
        add_greeting_column = add_greeting_column_func(file_input=iris_data)

    pipeline = spark_pipeline_from_builder(
        iris_data=iris_data,
    )
    return pipeline


def generate_dsl_pipeline_from_builder_with_dynamic_allocation_disabled() -> PipelineJob:
    # define the spark task
    iris_data = Input(path=parent_dir + "/dataset/iris.csv", type=AssetTypes.URI_FILE, mode=InputOutputModes.DIRECT)
    add_greeting_column_func = spark(
        code=parent_dir + "/src",
        entry={"file": "add_greeting_column.py"},
        py_files=["utils.zip"],
        files=["my_files.txt"],
        driver_cores=2,
        driver_memory="1g",
        executor_cores=1,
        executor_memory="1g",
        executor_instances=1,
        dynamic_allocation_min_executors=1,
        inputs=dict(file_input=iris_data),
        args="--file_input ${{inputs.file_input}}",
        resources={"instance_type": "standard_e4s_v3", "runtime_version": "3.4.0"},
    )

    # Define pipeline
    @dsl.pipeline(description="submit a pipeline with spark job")
    def spark_pipeline_from_builder(iris_data):
        add_greeting_column = add_greeting_column_func(file_input=iris_data)

    pipeline = spark_pipeline_from_builder(
        iris_data=iris_data,
    )
    return pipeline



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from typing import Any, Optional, cast

from azure.ai.ml.exceptions import ErrorCategory, ErrorTarget, ValidationException
from azure.ai.ml._restclient.v2024_10_01_preview.models import (
    ModelProvider as RestModelProvider,
    FineTuningVertical as RestFineTuningVertical,
    UriFileJobInput,
    MLFlowModelJobInput,
)
from azure.ai.ml.constants._common import AssetTypes
from azure.ai.ml._utils.utils import camel_to_snake
from azure.ai.ml.entities._inputs_outputs import Input
from azure.ai.ml.entities._job.finetuning.finetuning_job import FineTuningJob

from azure.ai.ml._utils._experimental import experimental


@experimental
class FineTuningVertical(FineTuningJob):
    def __init__(
        self,
        *,
        task: str,
        model: Input,
        model_provider: Optional[str],
        training_data: Input,
        validation_data: Optional[Input] = None,
        **kwargs: Any,
    ) -> None:
        self._task = task
        self._model = model
        self._model_provider = model_provider
        self._training_data = training_data
        self._validation_data = validation_data
        super().__init__(**kwargs)

    @property
    def task(self) -> str:
        """Get finetuning task.

        :return: The type of task to run. Possible values include: "ChatCompletion"
                 "TextCompletion", "TextClassification", "QuestionAnswering","TextSummarization",
                 "TokenClassification", "TextTranslation", "ImageClassification", "ImageInstanceSegmentation",
                 "ImageObjectDetection","VideoMultiObjectTracking".

        :rtype: str
        """
        return self._task

    @task.setter
    def task(self, task: str) -> None:
        """Set finetuning task.

        :param task: The type of task to run. Possible values include: "ChatCompletion"
                 "TextCompletion", "TextClassification", "QuestionAnswering","TextSummarization",
                 "TokenClassification", "TextTranslation", "ImageClassification", "ImageInstanceSegmentation",
                 "ImageObjectDetection","VideoMultiObjectTracking",.
        :type task: str

        :return: None
        """
        self._task = task

    @property
    def model(self) -> Optional[Input]:
        """The model to be fine-tuned.
        :return: Input object representing the mlflow model to be fine-tuned.
        :rtype: Input
        """
        return self._model

    @model.setter
    def model(self, value: Input) -> None:
        """Set the model to be fine-tuned.

        :param value: Input object representing the mlflow model to be fine-tuned.
        :type value: Input
        :raises ValidationException: Expected a mlflow model input.
        """
        if isinstance(value, Input) and (cast(Input, value).type in ("mlflow_model", "custom_model")):
            self._model = value
        else:
            msg = "Expected a mlflow model input or custom model input."
            raise ValidationException(
                message=msg,
                no_personal_data_message=msg,
                target=ErrorTarget.FINETUNING,
                error_category=ErrorCategory.USER_ERROR,
            )

    @property
    def model_provider(self) -> Optional[str]:
        """The model provider.
        :return: The model provider.
        :rtype: str
        """
        return self._model_provider

    @model_provider.setter
    def model_provider(self, value: str) -> None:
        """Set the model provider.

        :param value: The model provider.
        :type value: str
        """
        self._model_provider = RestModelProvider[camel_to_snake(value).upper()] if value else None

    @property
    def training_data(self) -> Input:
        """Get training data.

        :return: Training data input
        :rtype: Input
        """
        return self._training_data

    @training_data.setter
    def training_data(self, training_data: Input) -> None:
        """Set training data.

        :param training_data: Training data input
        :type training_data: Input
        """
        self._training_data = training_data

    @property
    def validation_data(self) -> Optional[Input]:
        """Get validation data.

        :return: Validation data input
        :rtype: Input
        """
        return self._validation_data

    @validation_data.setter
    def validation_data(self, validation_data: Input) -> None:
        """Set validation data.

        :param validation_data: Validation data input
        :type validation_data: Input
        """
        self._validation_data = validation_data

    def _resolve_inputs(self, rest_job: RestFineTuningVertical) -> None:
        """Resolve JobInputs to UriFileJobInput within data_settings.

        :param rest_job: The rest job object.
        :type rest_job: RestFineTuningVertical
        """
        if isinstance(rest_job.training_data, Input):
            rest_job.training_data = UriFileJobInput(uri=rest_job.training_data.path)
        if isinstance(rest_job.validation_data, Input):
            rest_job.validation_data = UriFileJobInput(uri=rest_job.validation_data.path)
        if isinstance(rest_job.model, Input):
            rest_job.model = MLFlowModelJobInput(uri=rest_job.model.path)

    def _restore_inputs(self) -> None:
        """Restore UriFileJobInputs to JobInputs within data_settings."""
        if isinstance(self.training_data, UriFileJobInput):
            self.training_data = Input(type=AssetTypes.URI_FILE, path=self.training_data.uri)
        if isinstance(self.validation_data, UriFileJobInput):
            self.validation_data = Input(type=AssetTypes.URI_FILE, path=self.validation_data.uri)
        if isinstance(self.model, MLFlowModelJobInput):
            self.model = Input(type=AssetTypes.MLFLOW_MODEL, path=self.model.uri)

    def __eq__(self, other: object) -> bool:
        """Returns True if both instances have the same values.

        This method check instances equality and returns True if both of
            the instances have the same attributes with the same values.

        :param other: Any object
        :type other: object
        :return: True or False
        :rtype: bool
        """
        if not isinstance(other, FineTuningVertical):
            return NotImplemented

        return (
            # TODO: Equality from base class does not work, no current precedence for this
            super().__eq__(other)
            and self.task == other.task
            and self.model == other.model
            and self.model_provider == other.model_provider
            and self.training_data == other.training_data
            and self.validation_data == other.validation_data
        )

    def __ne__(self, other: object) -> bool:
        """Check inequality between two FineTuningJob objects.

        :param other: Any object
        :type other: object
        :return: True or False
        :rtype: bool
        """
        return not self.__eq__(other)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from typing import Callable

import pytest
from devtools_testutils import AzureRecordedTestCase, is_live

from azure.ai.ml import MLClient
from azure.ai.ml.constants._workspace import IsolationMode
from azure.ai.ml.entities import FeatureStore, ManagedNetwork, PrivateEndpointDestination
from azure.core.polling import LROPoller


@pytest.mark.e2etest
@pytest.mark.data_experiences_test
@pytest.mark.usefixtures(
    "mock_workspace_arm_template_deployment_name",
    "mock_workspace_dependent_resource_name_generator",
)
class TestFeatureStore(AzureRecordedTestCase):
    @pytest.mark.skipif(
        condition=not is_live(),
        reason="ARM template makes playback complex, so the test is flaky when run against recording",
    )
    @pytest.mark.nofixdeploymentname
    @pytest.mark.nofixresourcename
    def test_feature_store_create_and_delete(
        self,
        client: MLClient,
        randstr: Callable[[], str],
        location: str,
        online_store_target: str,
        offline_store_target: str,
        materialization_identity_resource_id: str,
        materialization_identity_client_id: str,
        default_storage_account: str,
    ) -> None:
        from azure.ai.ml.entities import ManagedIdentityConfiguration, MaterializationStore

        if not online_store_target:
            print("Feature store create e2e test: online store target is not provided, use default target")
            online_store_target = "/subscriptions/1aefdc5e-3a7c-4d71-a9f9-f5d3b03be19a/resourceGroups/mdctest/providers/Microsoft.Cache/Redis/runhli-test"

        if not offline_store_target:
            print("Feature store create e2e test: offline store target is not provided, use default target")

        if not materialization_identity_client_id or not materialization_identity_resource_id:
            print("Feature store create e2e test: materialization identity is not provided, use default identity")

        if not default_storage_account:
            print("Feature store ceate e2e test: storage account is not provided, use default storage account")

        fs_name = f"e2etest_fs_{randstr('fs')}"
        fs_description = f"{fs_name} description"
        fs_display_name = f"{fs_name} display name"

        # provision with no store and identity
        fs = FeatureStore(name=fs_name, description=fs_description, display_name=fs_display_name, location=location)
        fs_poller = client.feature_stores.begin_create(feature_store=fs)
        assert isinstance(fs_poller, LROPoller)

        create_fs = fs_poller.result()
        assert isinstance(create_fs, FeatureStore)
        assert create_fs.materialization_identity.resource_id
        assert not create_fs._feature_store_settings.online_store_connection_name
        assert create_fs._feature_store_settings.offline_store_connection_name
        assert create_fs.offline_store.target
        assert create_fs.offline_store.target.startswith(create_fs.storage_account)

        if not materialization_identity_client_id or not materialization_identity_resource_id:
            materialization_identity_client_id = create_fs.materialization_identity.client_id
            materialization_identity_resource_id = create_fs.materialization_identity.resource_id

        if not offline_store_target:
            offline_store_target = create_fs.offline_store.target

        if not default_storage_account:
            default_storage_account = create_fs.storage_account

        # provision with online store
        fs_name = f"e2etest_fs1_{randstr('fs1')}"
        fs_description = f"{fs_name} description"
        fs_display_name = f"{fs_name} display name"

        fs1 = FeatureStore(
            name=fs_name,
            online_store=MaterializationStore(type="redis", target=online_store_target),
            description=fs_description,
            display_name=fs_display_name,
            location=location,
        )
        fs_poller = client.feature_stores.begin_create(feature_store=fs1)

        create_fs = fs_poller.result()
        assert isinstance(create_fs, FeatureStore)
        assert create_fs.materialization_identity.resource_id
        assert create_fs._feature_store_settings.online_store_connection_name
        assert create_fs._feature_store_settings.offline_store_connection_name
        assert create_fs.offline_store.target
        assert create_fs.online_store.target == online_store_target

        fs_poller = client.feature_stores.begin_delete(name=fs1.name, delete_dependent_resources=True)
        fs_poller.result()

        # provisioned with offline store
        fs_name = f"e2etest_fs2_{randstr('fs2')}"
        fs_description = f"{fs_name} description"
        fs_display_name = f"{fs_name} display name"

        fs2 = FeatureStore(
            name=fs_name,
            offline_store=MaterializationStore(type="azure_data_lake_gen2", target=offline_store_target),
            description=fs_description,
            display_name=fs_display_name,
            location=location,
        )
        fs_poller = client.feature_stores.begin_create(feature_store=fs2)

        create_fs = fs_poller.result()
        assert isinstance(create_fs, FeatureStore)
        assert create_fs.materialization_identity.resource_id
        assert not create_fs._feature_store_settings.online_store_connection_name
        assert create_fs._feature_store_settings.offline_store_connection_name
        assert create_fs.offline_store.target == offline_store_target
        assert not offline_store_target.startswith(create_fs.storage_account)

        fs_poller = client.feature_stores.begin_delete(name=fs2.name, delete_dependent_resources=True)
        fs_poller.result()

        # provisioned with materialization identity
        fs_name = f"e2etest_fs3_{randstr('fs3')}"
        fs_description = f"{fs_name} description"
        fs_display_name = f"{fs_name} display name"

        fs3 = FeatureStore(
            name=fs_name,
            materialization_identity=ManagedIdentityConfiguration(
                client_id=materialization_identity_client_id,
                resource_id=materialization_identity_resource_id,
            ),
            description=fs_description,
            display_name=fs_display_name,
            location=location,
        )
        fs_poller = client.feature_stores.begin_create(feature_store=fs3)

        create_fs = fs_poller.result()
        assert isinstance(create_fs, FeatureStore)
        assert create_fs.materialization_identity.resource_id.lower() == materialization_identity_resource_id.lower()
        assert not create_fs._feature_store_settings.online_store_connection_name
        assert create_fs._feature_store_settings.offline_store_connection_name
        assert create_fs.offline_store.target.startswith(create_fs.storage_account)

        fs_poller = client.feature_stores.begin_delete(name=fs3.name, delete_dependent_resources=True)
        fs_poller.result()

        fs_name = f"e2etest_fs4_{randstr('fs4')}"
        fs_description = f"{fs_name} description"
        fs_display_name = f"{fs_name} display name"

        fs4 = FeatureStore(
            name=fs_name,
            materialization_identity=ManagedIdentityConfiguration(
                client_id=materialization_identity_client_id,
                resource_id=materialization_identity_resource_id,
            ),
            offline_store=MaterializationStore(type="azure_data_lake_gen2", target=offline_store_target),
            online_store=MaterializationStore(type="redis", target=online_store_target),
            description=fs_description,
            display_name=fs_display_name,
            location=location,
        )
        fs_poller = client.feature_stores.begin_create(feature_store=fs4)

        create_fs = fs_poller.result()
        assert isinstance(create_fs, FeatureStore)
        assert create_fs.materialization_identity.resource_id.lower() == materialization_identity_resource_id.lower()
        assert create_fs._feature_store_settings.online_store_connection_name
        assert create_fs._feature_store_settings.offline_store_connection_name
        assert create_fs.online_store.target == online_store_target
        assert create_fs.offline_store.target == offline_store_target

        fs_poller = client.feature_stores.begin_delete(name=fs4.name, delete_dependent_resources=True)
        fs_poller.result()

        # provision with storage account
        fs_name = f"e2etest_fs5_{randstr('fs5')}"
        fs_description = f"{fs_name} description"
        fs_display_name = f"{fs_name} display name"

        fs5 = FeatureStore(
            name=fs_name,
            materialization_identity=ManagedIdentityConfiguration(
                client_id=materialization_identity_client_id,
                resource_id=materialization_identity_resource_id,
            ),
            storage_account=default_storage_account,
            offline_store=MaterializationStore(type="azure_data_lake_gen2", target=offline_store_target),
            description=fs_description,
            display_name=fs_display_name,
            location=location,
        )
        fs_poller = client.feature_stores.begin_create(feature_store=fs5)

        create_fs = fs_poller.result()
        assert create_fs.materialization_identity.resource_id.lower() == materialization_identity_resource_id.lower()
        assert create_fs.storage_account == default_storage_account
        assert not create_fs._feature_store_settings.online_store_connection_name
        assert create_fs._feature_store_settings.offline_store_connection_name
        assert create_fs.offline_store.target == offline_store_target

        fs_name = f"e2etest_fs6_{randstr('fs6')}"
        fs_description = f"{fs_name} description"
        fs_display_name = f"{fs_name} display name"

        fs6 = FeatureStore(
            name=fs_name,
            materialization_identity=ManagedIdentityConfiguration(
                client_id=materialization_identity_client_id,
                resource_id=materialization_identity_resource_id,
            ),
            storage_account=default_storage_account,
            description=fs_description,
            display_name=fs_display_name,
            location=location,
        )
        fs_poller = client.feature_stores.begin_create(feature_store=fs6)

        create_fs = fs_poller.result()
        assert create_fs.materialization_identity.resource_id.lower() == materialization_identity_resource_id.lower()
        assert create_fs.storage_account == default_storage_account
        assert not create_fs._feature_store_settings.online_store_connection_name
        assert create_fs._feature_store_settings.offline_store_connection_name
        assert create_fs.offline_store.target
        assert not create_fs.offline_store.target.startswith(default_storage_account)

        fs_poller = client.feature_stores.begin_delete(name=fs5.name, delete_dependent_resources=True)
        fs_poller.result()
        fs_poller = client.feature_stores.begin_delete(name=fs6.name, delete_dependent_resources=True)
        fs_poller.result()
        fs_poller = client.feature_stores.begin_delete(name=fs.name, delete_dependent_resources=True)
        fs_poller.result()

        # delete test
        with pytest.raises(Exception):
            client.feature_stores.get(fs.name)

        # normal workspace create
        from azure.ai.ml.entities import Workspace

        ws_name = f"e2etest_ws_{randstr('ws')}"

        ws = Workspace(name=ws_name, location=location)
        ws_poller = client.workspaces.begin_create(workspace=ws)
        create_ws = ws_poller.result()
        provisioned_storage_account = create_ws.storage_account

        assert isinstance(create_ws, Workspace)
        assert create_ws.name == ws_name

        ws1_name = f"e2etest_ws1_{randstr('ws1')}"
        ws1 = Workspace(name=ws1_name, storage_account=provisioned_storage_account, location=location)
        ws_poller = client.workspaces.begin_create(workspace=ws1)
        create_ws = ws_poller.result()

        assert isinstance(create_ws, Workspace)
        assert create_ws.name == ws1_name
        assert create_ws.storage_account == provisioned_storage_account

        ws_poller = client.workspaces.begin_delete(
            name=ws.name, delete_dependent_resources=True, permanently_delete=True
        )
        ws_poller.result()
        ws_poller = client.workspaces.begin_delete(
            name=ws1.name, delete_dependent_resources=True, permanently_delete=True
        )
        ws_poller.result()

    @pytest.mark.skipif(
        condition=not is_live(),
        reason="ARM template makes playback complex, so the test is flaky when run against recording",
    )
    @pytest.mark.nofixdeploymentname
    @pytest.mark.nofixresourcename
    def test_feature_store_update_delete(
        self,
        client: MLClient,
        randstr: Callable[[], str],
        location: str,
        online_store_target: str,
        offline_store_target: str,
        materialization_identity_resource_id: str,
        materialization_identity_client_id: str,
    ) -> None:
        if not online_store_target:
            print("Feature store update and delete e2e test: online store target is not provided, use default target")
            online_store_target = "/subscriptions/1aefdc5e-3a7c-4d71-a9f9-f5d3b03be19a/resourceGroups/mdctest/providers/Microsoft.Cache/Redis/runhli-test"

        if not offline_store_target:
            print("Feature store update and delete e2e test: offline store target is not provided, use default target")

        if not materialization_identity_client_id or not materialization_identity_resource_id:
            print(
                "Feature store update and delete e2e test: materialization identity is not provided, use default identity"
            )

        alt_fs_name = None
        if (
            not offline_store_target
            or not materialization_identity_client_id
            or not materialization_identity_resource_id
        ):
            alt_fs_name = f"alt_fs_{randstr('alt_fs')}"
            alt_fs_description = f"{alt_fs_name} description"
            alt_fs = FeatureStore(name=alt_fs_name, description=alt_fs_description, location=location)
            fs_poller = client.feature_stores.begin_create(feature_store=alt_fs)
            create_fs = fs_poller.result()
            offline_store_target = create_fs.offline_store.target
            materialization_identity_client_id = create_fs.materialization_identity.client_id
            materialization_identity_resource_id = create_fs.materialization_identity.resource_id

        fs_name = f"e2etest_fs_{randstr('fs_to_update')}"
        fs_description = f"{fs_name} description"
        fs_updated_description = f"{fs_name} updated description"
        fs_display_name = f"{fs_name} display name"

        fs = FeatureStore(name=fs_name, description=fs_description, display_name=fs_display_name, location=location)
        fs_poller = client.feature_stores.begin_create(feature_store=fs)

        create_fs = fs_poller.result()
        assert isinstance(create_fs, FeatureStore)

        assert create_fs.materialization_identity.resource_id
        assert create_fs.offline_store.target
        assert not create_fs._feature_store_settings.online_store_connection_name
        assert create_fs._feature_store_settings.offline_store_connection_name
        provisioned_offline_store_resource_id = create_fs.offline_store.target
        provisioned_materialization_identity = create_fs.materialization_identity

        fs_result = client.feature_stores.get(fs_name)
        assert isinstance(fs_result, FeatureStore)
        assert fs_result.name == fs_name

        fs_list = client.feature_stores.list()
        assert any(fs_name == fs.name for fs in fs_list)

        updated_fs = FeatureStore(name=fs_name, description=fs_updated_description, display_name=fs_display_name)
        fs_poller = client.feature_stores.begin_update(feature_store=updated_fs)

        updated_fs = fs_poller.result()
        assert isinstance(updated_fs, FeatureStore)
        assert not updated_fs._feature_store_settings.online_store_connection_name
        assert updated_fs._feature_store_settings.offline_store_connection_name
        assert updated_fs.name == fs_name
        assert updated_fs.description == fs_updated_description

        from azure.ai.ml.entities import ManagedIdentityConfiguration, MaterializationStore

        # online store update
        online_fs = FeatureStore(
            name=fs_name,
            online_store=MaterializationStore(type="redis", target=online_store_target),
        )
        fs_poller = client.feature_stores.begin_update(feature_store=online_fs)

        online_fs = fs_poller.result()
        assert isinstance(online_fs, FeatureStore)
        assert online_fs._feature_store_settings.online_store_connection_name
        assert online_fs._feature_store_settings.offline_store_connection_name
        assert online_fs.materialization_identity.resource_id == provisioned_materialization_identity.resource_id
        assert online_fs.offline_store.target == provisioned_offline_store_resource_id
        assert online_fs.online_store.target == online_store_target

        # offline store update
        offline_fs = FeatureStore(
            name=fs_name,
            offline_store=MaterializationStore(type="azure_data_lake_gen2", target=offline_store_target),
        )
        fs_poller = client.feature_stores.begin_update(feature_store=offline_fs)
        offline_fs = fs_poller.result()
        assert isinstance(offline_fs, FeatureStore)
        assert offline_fs.materialization_identity.resource_id == provisioned_materialization_identity.resource_id
        assert offline_fs._feature_store_settings.online_store_connection_name
        assert offline_fs._feature_store_settings.offline_store_connection_name
        assert (
            offline_fs._feature_store_settings.offline_store_connection_name
            != online_fs._feature_store_settings.offline_store_connection_name
        )
        assert offline_fs.offline_store.target == offline_store_target
        assert online_fs.online_store.target == online_store_target

        # identity update
        identity_fs = FeatureStore(
            name=fs_name,
            materialization_identity=ManagedIdentityConfiguration(
                client_id=materialization_identity_client_id, resource_id=materialization_identity_resource_id
            ),
        )
        fs_poller = client.feature_stores.begin_update(feature_store=identity_fs)
        identity_fs = fs_poller.result()
        assert isinstance(identity_fs, FeatureStore)
        assert identity_fs.materialization_identity.client_id == materialization_identity_client_id
        assert identity_fs.materialization_identity.resource_id.lower() == materialization_identity_resource_id.lower()
        assert (
            identity_fs._feature_store_settings.online_store_connection_name
            == online_fs._feature_store_settings.online_store_connection_name
        )
        assert (
            identity_fs._feature_store_settings.offline_store_connection_name
            == offline_fs._feature_store_settings.offline_store_connection_name
        )

        identity_fs = FeatureStore(
            name=fs_name,
            materialization_identity=ManagedIdentityConfiguration(
                client_id=provisioned_materialization_identity.client_id,
                resource_id=provisioned_materialization_identity.resource_id,
            ),
            offline_store=MaterializationStore(type="azure_data_lake_gen2", target=offline_store_target),
        )
        fs_poller = client.feature_stores.begin_update(feature_store=identity_fs)
        identity_fs = fs_poller.result()
        assert isinstance(offline_fs, FeatureStore)
        assert identity_fs.materialization_identity.client_id == provisioned_materialization_identity.client_id
        assert (
            identity_fs.materialization_identity.resource_id.lower()
            == provisioned_materialization_identity.resource_id.lower()
        )
        assert identity_fs.offline_store.target == offline_store_target
        assert identity_fs.online_store.target == online_store_target
        assert (
            identity_fs._feature_store_settings.online_store_connection_name
            == online_fs._feature_store_settings.online_store_connection_name
        )
        assert (
            identity_fs._feature_store_settings.offline_store_connection_name
            == offline_fs._feature_store_settings.offline_store_connection_name
        )

        # test update and provision managed network interface
        managed_network_fs = FeatureStore(
            name=fs_name,
            public_network_access="disabled",
            managed_network=ManagedNetwork(
                isolation_mode=IsolationMode.ALLOW_INTERNET_OUTBOUND,
                outbound_rules=[
                    PrivateEndpointDestination(
                        name="sourcerulefs",
                        service_resource_id=create_fs.storage_account,
                        subresource_target="dfs",
                        spark_enabled="true",
                    ),
                    PrivateEndpointDestination(
                        name="defaultkeyvault",
                        service_resource_id=create_fs.key_vault,
                        subresource_target="vault",
                        spark_enabled="true",
                    ),
                ],
            ),
        )
        fs_poller = client.feature_stores.begin_update(feature_store=managed_network_fs)
        fs_poller.result()
        status_poller = client.feature_stores.begin_provision_network(feature_store_name=fs_name)
        assert isinstance(status_poller, LROPoller)

        fs_poller = client.feature_stores.begin_delete(name=fs.name, delete_dependent_resources=True)
        if alt_fs_name:
            fs_poller = client.feature_stores.begin_delete(name=alt_fs_name, delete_dependent_resources=True)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

from marshmallow import fields
from azure.ai.ml._schema.core.schema import PatchedSchemaMeta
from azure.ai.ml._utils._experimental import experimental


@experimental
class AzureOpenAIHyperparametersSchema(metaclass=PatchedSchemaMeta):
    n_epochs = fields.Int()
    learning_rate_multiplier = fields.Float()
    batch_size = fields.Int()
    # TODO: Should be dict<string,string>, check schema for the same.
    # For now not exposing as we dont have REST layer representation exposed.
    # Need to check with the team.
    # additional_parameters = fields.Dict()



import argparse
from random import random

from azureml.core import Run


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--lr", required=True)
    parser.add_argument("--conv_size", required=True)
    parser.add_argument("--dropout_rate", required=True)

    args = parser.parse_args()
    print("validated")
    print(args.lr)
    print(args.conv_size)
    print(args.dropout_rate)

    run = Run.get_context()
    run.log("accuracy", random())


if __name__ == "__main__":
    main()



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import json
import logging
import traceback
from typing import Dict, NoReturn, Optional, Tuple, Union

from colorama import Fore, Style, init
from marshmallow.exceptions import ValidationError as SchemaValidationError

from azure.ai.ml.constants._common import (
    DATASTORE_SCHEMA_TYPES,
    SCHEMA_VALIDATION_ERROR_TEMPLATE,
    YAML_CREATION_ERROR_DESCRIPTION,
)
from azure.ai.ml.exceptions import ErrorTarget, MlException, ValidationErrorType, ValidationException

module_logger = logging.getLogger(__name__)


def _find_deepest_dictionary(data: dict) -> dict:
    """Find deepest dictionary in nested dictionary.

    Used here to get nested error message. Can't be in utils.py due to circular import.

    :param data: The dictionary
    :type data: dict
    :return: The deepest dictionary
    :rtype: dict
    """

    for v in data.values():
        if isinstance(v, dict):
            return _find_deepest_dictionary(v)

    return data


def get_entity_type(error: Union[str, SchemaValidationError, ValidationException]) -> Tuple[str, str]:
    """Get entity name from schema type referenced in the error.

    :param error: The validation error
    :type error: Union[str, SchemaValidationError, ValidationException]
    :return: The entity type and additional details
    :rtype: Tuple[str, str]
    """
    details = ""

    error_name = error.exc_msg if hasattr(error, "exc_msg") else error.split(":")[0]
    if isinstance(error, ValidationException):
        entity_type = error.target
    else:
        details += error_name + "\n"
        if "DataSchema" in error_name:
            entity_type = ErrorTarget.DATA
        elif "ModelSchema" in error_name:
            entity_type = ErrorTarget.MODEL
        elif "EnvironmentSchema" in error_name:
            entity_type = ErrorTarget.ENVIRONMENT
        elif "CodeAssetSchema" in error_name:
            entity_type = ErrorTarget.CODE
        elif any(x in error_name for x in DATASTORE_SCHEMA_TYPES):
            entity_type = ErrorTarget.DATASTORE
        elif "BaseJobSchema" in error_name:
            entity_type = ErrorTarget.JOB
        elif "CommandSchema" in error_name:
            entity_type = ErrorTarget.COMMAND_JOB
        elif "CommandJobSchema" in error_name:
            entity_type = ErrorTarget.COMMAND_JOB
        elif "SweepSchema" in error_name:
            entity_type = ErrorTarget.SWEEP_JOB
        elif "SweepJobSchema" in error_name:
            entity_type = ErrorTarget.SWEEP_JOB
        elif "AutoMLJobSchema" in error_name:
            entity_type = ErrorTarget.AUTOML
        else:
            raise NotImplementedError()

    return entity_type, details


def format_details_section(
    error: Union[str, SchemaValidationError, ValidationException], details: str, entity_type: str
) -> Tuple[Dict[str, bool], str]:
    """Builds strings for details of the error message template's Details section.

    :param error: The validation error
    :type error: Union[str, SchemaValidationError, ValidationException]
    :param details: The details
    :type details: str
    :param entity_type: The entity type
    :type entity_type: str
    :return: Error type map and formatted message
    :rtype: Tuple[Dict[str, bool], str]
    """

    error_types = {
        ValidationErrorType.INVALID_VALUE: False,
        ValidationErrorType.UNKNOWN_FIELD: False,
        ValidationErrorType.MISSING_FIELD: False,
        ValidationErrorType.FILE_OR_FOLDER_NOT_FOUND: False,
        ValidationErrorType.CANNOT_SERIALIZE: False,
        ValidationErrorType.CANNOT_PARSE: False,
        ValidationErrorType.RESOURCE_NOT_FOUND: False,
    }

    if hasattr(error, "message"):
        error_types[error.error_type] = True
        details += f"\n\n{Fore.RED}(x) {error.message}{Fore.RESET}\n"
    else:
        if (
            entity_type == ErrorTarget.COMMAND_JOB
        ):  # Command Job errors require different parsing because they're using ValidationResult.
            # This separate treatment will be unnecessary once
            # task https://msdata.visualstudio.com/Vienna/_workitems/edit/1925982/ is resolved.
            parsed_error_msg = error[error.find("{") : (error.rfind("}") + 1)]
            error_msg = json.loads(parsed_error_msg).get("errors")[0].get("message")
            error_msg = error_msg[error_msg.find("{") : (error_msg.rfind("}") + 1)]
            error_msg = error_msg[: (error_msg.rfind("}") + 1)]
            error_msg = json.loads(error_msg)
        else:
            parsed_error_msg = error[error.find("{") : (error.rfind("}") + 1)]
            error_msg = json.loads(parsed_error_msg)

        for field, field_error in zip(error_msg.keys(), error_msg.values()):
            try:
                field_error = field_error[0]
            except KeyError:
                error_msg = _find_deepest_dictionary(field_error)
                if entity_type not in [ErrorTarget.PIPELINE, ErrorTarget.COMMAND_JOB]:
                    field = list(error_msg.keys())[0]
                field_error = list(error_msg.values())[0][0]
            field_error_string = str(field_error)

            if not error_types[ValidationErrorType.INVALID_VALUE] and any(
                s in field_error_string for s in ["Not a valid", "is not in set"]
            ):
                error_types[ValidationErrorType.INVALID_VALUE] = True
            if not error_types[ValidationErrorType.UNKNOWN_FIELD] and "Unknown field" in field_error_string:
                error_types[ValidationErrorType.UNKNOWN_FIELD] = True
            if (
                not error_types[ValidationErrorType.MISSING_FIELD]
                and "Missing data for required field" in field_error_string
            ):
                error_types[ValidationErrorType.MISSING_FIELD] = True
            if (
                not error_types[ValidationErrorType.FILE_OR_FOLDER_NOT_FOUND]
                and "No such file or directory" in field_error_string
            ):
                error_types[ValidationErrorType.FILE_OR_FOLDER_NOT_FOUND] = True
            if not error_types[ValidationErrorType.CANNOT_SERIALIZE] and "Cannot dump" in field_error_string:
                error_types[ValidationErrorType.CANNOT_SERIALIZE] = True
            if (
                not error_types[ValidationErrorType.CANNOT_PARSE]
                and "Error while parsing yaml file" in field_error_string
            ):
                error_types[ValidationErrorType.CANNOT_PARSE] = True

            if isinstance(field_error, dict):
                field_error = f"{list(field_error.keys())[0]}:\n    - {list(field_error.values())[0][0]}"

            details += f"{Fore.RED}\n(x) {field}:\n- {field_error}{Fore.RESET}\n"
    return error_types, details


def format_errors_and_resolutions_sections(
    entity_type: str, error_types: Dict[str, bool], cli: bool
) -> Tuple[str, str]:
    """Builds strings for details of the error message template's Errors and Resolutions sections.

    :param entity_type: The entity type
    :type entity_type: str
    :param error_types: A dict that specifies which errors were encountered
    :type error_types: Dict[str, bool]
    :param cli: Whether this was invoked from Azure CLI AzureML extension.
    :type cli: bool
    :return: Error and Resolution messages
    :rtype: Tuple[str, str]
    """

    resolutions = ""
    errors = ""
    count = 1

    if error_types[ValidationErrorType.INVALID_VALUE]:
        errors += f"\n{count}) One or more fields are invalid"
        resolutions += (
            f"\n{count}) Double-check that all specified parameters are of the correct types and formats "
            f"prescribed by the {entity_type} schema."
        )
        count += 1
    if error_types[ValidationErrorType.UNKNOWN_FIELD]:
        errors += f"\n{count}) A least one unrecognized parameter is specified"
        resolutions += f"\n{count}) Remove any parameters not prescribed by the {entity_type} schema."
        count += 1
    if error_types[ValidationErrorType.MISSING_FIELD]:
        errors += f"\n{count}) At least one required parameter is missing"
        resolutions += f"\n{count}) Ensure all parameters required by the {entity_type} schema are specified."
        count += 1
    if error_types[ValidationErrorType.FILE_OR_FOLDER_NOT_FOUND]:
        errors += f"\n{count}) One or more files or folders do not exist.\n"
        resolutions += f"\n{count}) Double-check the directory path you provided and enter the correct path."
        count += 1
    if error_types[ValidationErrorType.CANNOT_SERIALIZE]:
        errors += f"\n{count}) One or more fields cannot be serialized.\n"
        resolutions += f"\n{count}) Double-check that all specified parameters are of the correct types and formats \
        prescribed by the {entity_type} schema."
        count += 1
    if error_types[ValidationErrorType.CANNOT_PARSE]:
        errors += f"\n{count}) YAML file cannot be parsed.\n"
        resolutions += f"\n{count}) Double-check your YAML file for syntax and formatting errors."
        count += 1
    if error_types[ValidationErrorType.RESOURCE_NOT_FOUND]:
        errors += f"\n{count}) Resource was not found.\n"
        resolutions += (
            f"\n{count}) Double-check that the resource has been specified correctly and " "that you have access to it."
        )
        count += 1

    if cli:
        errors = "Error: " + errors
    else:
        errors = Fore.BLACK + errors + Fore.RESET

    return errors, resolutions


def format_create_validation_error(
    error: Union[SchemaValidationError, ValidationException],
    yaml_operation: bool,
    cli: bool = False,
    raw_error: Optional[str] = None,
) -> str:
    """Formats a detailed error message for validation errors.

    :param error: The validation error
    :type error: Union[SchemaValidationError, ValidationException]
    :param yaml_operation: Whether the exception was caused by yaml validation.
    :type yaml_operation: bool
    :param cli: Whether is invoked from Azure CLI AzureML Extension. Defaults to False
    :type cli: bool
    :param raw_error: The raw exception message
    :type raw_error: Optional[str]
    :return: Formatted error message
    :rtype: str
    """
    from azure.ai.ml._schema._datastore import (
        AzureBlobSchema,
        AzureDataLakeGen1Schema,
        AzureDataLakeGen2Schema,
        AzureFileSchema,
    )
    from azure.ai.ml._schema._sweep import SweepJobSchema
    from azure.ai.ml._schema.assets.data import DataSchema
    from azure.ai.ml._schema.assets.environment import EnvironmentSchema
    from azure.ai.ml._schema.assets.model import ModelSchema
    from azure.ai.ml._schema.job import CommandJobSchema
    from azure.ai.ml.entities._util import REF_DOC_ERROR_MESSAGE_MAP

    if raw_error:
        error = raw_error
    entity_type, details = get_entity_type(error)
    error_types, details = format_details_section(error, details, entity_type)
    errors, resolutions = format_errors_and_resolutions_sections(entity_type, error_types, cli)

    if yaml_operation:
        description = YAML_CREATION_ERROR_DESCRIPTION.format(entity_type=entity_type)
        description = Style.BRIGHT + description + Style.RESET_ALL

        if entity_type == ErrorTarget.MODEL:
            schema_type = ModelSchema
        elif entity_type == ErrorTarget.DATA:
            schema_type = DataSchema
        elif entity_type == ErrorTarget.COMMAND_JOB:
            schema_type = CommandJobSchema
        elif entity_type == ErrorTarget.SWEEP_JOB:
            schema_type = SweepJobSchema
        elif entity_type in [ErrorTarget.BLOB_DATASTORE, ErrorTarget.DATASTORE]:
            schema_type = AzureBlobSchema
        elif entity_type == ErrorTarget.GEN1_DATASTORE:
            schema_type = AzureDataLakeGen1Schema
        elif entity_type == ErrorTarget.GEN2_DATASTORE:
            schema_type = AzureDataLakeGen2Schema
        elif entity_type == ErrorTarget.FILE_DATASTORE:
            schema_type = AzureFileSchema
        elif entity_type == ErrorTarget.ENVIRONMENT:
            schema_type = EnvironmentSchema
        else:
            schema_type = ""

        resolutions += " " + REF_DOC_ERROR_MESSAGE_MAP.get(schema_type, "")
    else:
        description = ""

    resolutions += "\n"
    formatted_error = SCHEMA_VALIDATION_ERROR_TEMPLATE.format(
        description=description,
        error_msg=errors,
        parsed_error_details=details,
        resolutions=resolutions,
        text_color=Fore.WHITE,
        link_color=Fore.CYAN,
        reset=Fore.RESET,
    )

    return formatted_error


def log_and_raise_error(error: Exception, debug: bool = False, yaml_operation: bool = False) -> NoReturn:
    init()

    # use an f-string to automatically call str() on error
    if debug:
        module_logger.error(traceback.print_exc())

    if isinstance(error, SchemaValidationError):
        module_logger.debug(traceback.format_exc())
        try:
            formatted_error = format_create_validation_error(error.messages[0], yaml_operation=yaml_operation)
        except NotImplementedError:
            formatted_error = error
    elif isinstance(error, ValidationException):
        module_logger.debug(traceback.format_exc())
        try:
            error_type = error.error_type
            if error_type == ValidationErrorType.GENERIC:
                formatted_error = error
            else:
                formatted_error = format_create_validation_error(error, yaml_operation=yaml_operation)
        except NotImplementedError:
            formatted_error = error
    else:
        raise error

    raise MlException(message=formatted_error, no_personal_data_message=formatted_error)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import pytest
import json

from azure.ai.ml.entities._assets._artifacts.feature_set import FeatureSet
from azure.ai.ml.entities._load_functions import load_feature_set
from azure.ai.ml._restclient.v2023_10_01.models import (
    FeaturesetContainer,
    FeaturesetContainerProperties,
    FeaturesetVersion,
    FeaturesetVersionProperties,
)


@pytest.mark.unittest
class TestFeatureSetEntity:
    FEATURE_SET = "./tests/test_configs/feature_set/feature_set_full.yaml"
    FEATURE_SET_REST = "./tests/test_configs/feature_set/feature_set_full_rest.json"
    FEATURE_SET_CONTAINER_REST = "./tests/test_configs/feature_set/feature_set_container_rest.json"

    def test_to_rest_object(self) -> None:
        feature_set = load_feature_set(self.FEATURE_SET)

        feature_set_rest = feature_set._to_rest_object()
        assert feature_set_rest.properties.description == feature_set.description
        assert feature_set_rest.properties.tags == feature_set.tags
        assert feature_set_rest.properties.entities == feature_set.entities
        assert feature_set_rest.properties.specification.path == feature_set.specification.path
        assert feature_set_rest.properties.stage == feature_set.stage
        assert (
            str(feature_set_rest.properties.materialization_settings.store_type) == "MaterializationStoreType.OFFLINE"
        )
        assert (
            feature_set_rest.properties.materialization_settings.schedule.frequency
            == feature_set.materialization_settings.schedule.frequency
        )
        assert (
            feature_set_rest.properties.materialization_settings.schedule.interval
            == feature_set.materialization_settings.schedule.interval
        )
        assert (
            feature_set_rest.properties.materialization_settings.spark_configuration
            == feature_set.materialization_settings.spark_configuration
        )
        assert (
            feature_set_rest.properties.materialization_settings.resource.instance_type
            == feature_set.materialization_settings.resource.instance_type
        )
        assert (
            feature_set_rest.properties.materialization_settings.notification.email_on
            == feature_set.materialization_settings.notification.email_on
        )
        assert (
            feature_set_rest.properties.materialization_settings.notification.emails
            == feature_set.materialization_settings.notification.emails
        )

    def test_from_rest_object(self) -> None:
        with open(self.FEATURE_SET_REST, "r") as f:
            feature_set_rest = FeaturesetVersion.deserialize(json.load(f))
            feature_set = FeatureSet._from_rest_object(featureset_rest_object=feature_set_rest)
            assert feature_set.name == feature_set_rest.name
            assert feature_set.description == feature_set_rest.properties.description
            assert feature_set.tags == feature_set_rest.properties.tags
            assert feature_set.entities == feature_set_rest.properties.entities
            assert feature_set.specification.path == feature_set_rest.properties.specification.path
            assert feature_set.stage == feature_set_rest.properties.stage
            assert (
                feature_set.materialization_settings.schedule.frequency
                == feature_set_rest.properties.materialization_settings.schedule.frequency.lower()
            )
            assert (
                feature_set.materialization_settings.schedule.interval
                == feature_set_rest.properties.materialization_settings.schedule.interval
            )
            assert (
                feature_set.materialization_settings.spark_configuration
                == feature_set_rest.properties.materialization_settings.spark_configuration
            )
            assert (
                feature_set.materialization_settings.resource.instance_type
                == feature_set_rest.properties.materialization_settings.resource.instance_type
            )
            assert (
                feature_set.materialization_settings.notification.email_on
                == feature_set_rest.properties.materialization_settings.notification.email_on
            )
            assert (
                feature_set.materialization_settings.notification.emails
                == feature_set_rest.properties.materialization_settings.notification.emails
            )

    def test_from_container_rest_object(self) -> None:
        with open(self.FEATURE_SET_CONTAINER_REST, "r") as f:
            featureset_container_rest = FeaturesetContainer.deserialize(json.load(f))
            feature_set = FeatureSet._from_container_rest_object(featureset_container_rest)
            assert feature_set.name == featureset_container_rest.name
            assert feature_set.description == featureset_container_rest.properties.description
            assert feature_set.tags == featureset_container_rest.properties.tags
            assert feature_set.entities == []
            assert feature_set.specification.path == None
            assert feature_set.version == ""
            assert feature_set.latest_version == featureset_container_rest.properties.latest_version



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
from dataclasses import dataclass
from typing import Any, Dict, Optional
from azure.ai.ml._utils._experimental import experimental
from azure.ai.ml._utils.utils import camel_to_snake
from azure.ai.ml.entities._workspace.connections.workspace_connection import WorkspaceConnection
from azure.ai.ml.entities._workspace.connections.connection_subtypes import (
    AzureOpenAIConnection,
    AadCredentialConfiguration,
)


@experimental
@dataclass
class ModelConfiguration:
    """Configuration for a embedding model.

    :param api_base: The base URL for the API.
    :type api_base: Optional[str]
    :param api_key: The API key.
    :type api_key: Optional[str]
    :param api_version: The API version.
    :type api_version: Optional[str]
    :param model_name: The name of the model.
    :type model_name: Optional[str]
    :param model_name: The deployment name of the model.
    :type model_name: Optional[str]
    :param connection_name: The name of the workspace connection of this model.
    :type connection_name: Optional[str]
    :param connection_type: The type of the workspace connection of this model.
    :type connection_type: Optional[str]
    :param model_kwargs: Additional keyword arguments for the model.
    :type model_kwargs: Dict[str, Any]
    """

    api_base: Optional[str]
    api_key: Optional[str]
    api_version: Optional[str]
    connection_name: Optional[str]
    connection_type: Optional[str]
    model_name: Optional[str]
    deployment_name: Optional[str]
    model_kwargs: Dict[str, Any]

    def __init__(
        self,
        *,
        api_base: Optional[str],
        api_key: Optional[str],
        api_version: Optional[str],
        connection_name: Optional[str],
        connection_type: Optional[str],
        model_name: Optional[str],
        deployment_name: Optional[str],
        model_kwargs: Dict[str, Any]
    ):
        self.api_base = api_base
        self.api_key = api_key
        self.api_version = api_version
        self.connection_name = connection_name
        self.connection_type = connection_type
        self.model_name = model_name
        self.deployment_name = deployment_name
        self.model_kwargs = model_kwargs

    @staticmethod
    def from_connection(
        connection: WorkspaceConnection,
        model_name: Optional[str] = None,
        deployment_name: Optional[str] = None,
        **kwargs
    ) -> "ModelConfiguration":
        """Create an model configuration from a Connection.

        :param connection: The WorkspaceConnection object.
        :type connection: ~azure.ai.ml.entities.WorkspaceConnection
        :param model_name: The name of the model.
        :type model_name: Optional[str]
        :param deployment_name: The name of the deployment.
        :type deployment_name: Optional[str]
        :return: The model configuration.
        :rtype: ~azure.ai.ml.entities._indexes.entities.ModelConfiguration
        :raises TypeError: If the connection is not an AzureOpenAIConnection.
        :raises ValueError: If the connection does not contain an OpenAI key.
        """
        if isinstance(connection, AzureOpenAIConnection) or camel_to_snake(connection.type) == "azure_open_ai":
            connection_type = "azure_open_ai"
            api_version = connection.api_version  # type: ignore[attr-defined]
            if not model_name or not deployment_name:
                raise ValueError("Please specify model_name and deployment_name.")
        elif connection.type and connection.type.lower() == "serverless":
            connection_type = "serverless"
            api_version = None
            if not connection.id:
                raise TypeError("The connection id is missing from the serverless connection object.")
        else:
            raise TypeError("Connection object is not supported.")

        if isinstance(connection.credentials, AadCredentialConfiguration):
            key = None
        else:
            key = connection.credentials.get("key")  # type: ignore[union-attr]
            if key is None and connection_type == "azure_open_ai":
                import os

                if "AZURE_OPENAI_API_KEY" in os.environ:
                    key = os.getenv("AZURE_OPENAI_API_KEY")
                else:
                    raise ValueError("Unable to retrieve openai key from connection object or env variable.")

        return ModelConfiguration(
            api_base=connection.target,
            api_key=key,
            api_version=api_version,
            connection_name=connection.name,
            connection_type=connection_type,
            model_name=model_name,
            deployment_name=deployment_name,
            model_kwargs=kwargs,
        )



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=protected-access

import http.client
import logging
import os
import sys
import time
import traceback
from collections import namedtuple
from typing import List, Optional

LOG_FILE = os.path.abspath("azureml.log")
LOG_FORMAT = "%(asctime)s|%(name)s|%(levelname)s|%(message)s"
INTERESTING_NAMESPACES = ["azureml", "msrest.http_logger", "urllib2", "azure"]

module_logger = logging.getLogger(__name__)
separator = "\n==================\n"
ConnectionInfo = namedtuple("ConnectionInfo", ["host", "port", "hasSocket"])


def stack_info() -> list:
    main_stack = []
    for threadId, stack in sys._current_frames().items():
        for filename, lineno, name, line in traceback.extract_stack(stack):
            call = line.strip() if line is not None else None
            main_stack.append(
                {
                    "ThreadID": threadId,
                    "File": filename,
                    "Line": lineno,
                    "Name": name,
                    "Call": call,
                }
            )

    return main_stack


def connection_info(gc_objects: list) -> List[ConnectionInfo]:
    connections = [obj for obj in gc_objects if isinstance(obj, http.client.HTTPConnection)]
    return [ConnectionInfo(host=c.host, port=c.port, hasSocket=c.sock is not None) for c in connections]  # disable


# pylint: disable=client-incorrect-naming-convention
class diagnostic_log(object):
    """Directs debug logs to a specified file.

    :param log_path: A path with log file name. If None, a file named "azureml.log" is
        created.
    :type log_path: str
    :param namespaces: A list of namespaces to capture logs for. If None, the default is "azureml",
        "msrest.http_logger", "urllib2", and "azure".
    :type namespaces: builtin.list
    :param context_name: A name to identify the logging context. If None, the context of the calling
        stack frame is used.
    :type context_type: str
    """

    def __init__(
        self, log_path: Optional[str] = None, namespaces: Optional[list] = None, context_name: Optional[str] = None
    ):
        self._namespaces = INTERESTING_NAMESPACES if namespaces is None else namespaces
        self._filename = LOG_FILE if log_path is None else log_path
        self._filename = os.path.abspath(self._filename)
        self._capturing = False
        if context_name is None:
            import inspect

            context_name = inspect.getouterframes(inspect.currentframe(), 2)[1].function
        self._context_name = context_name

        formatter = logging.Formatter(LOG_FORMAT)
        formatter.converter = time.gmtime

        file_handler = logging.FileHandler(filename=self._filename, encoding="utf-8")
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(formatter)
        self._handler = file_handler

    def start_capture(self) -> None:
        """Start the capture of debug logs."""
        if self._capturing:
            module_logger.warning("Debug logs are already enabled at %s", self._filename)
            return

        print("Debug logs are being sent to {}".format(self._filename))
        for namespace in self._namespaces:
            module_logger.debug("Adding [%s] debug logs to this file", namespace)
            n_logger = logging.getLogger(namespace)
            n_logger.setLevel(logging.DEBUG)
            n_logger.addHandler(self._handler)
            # We do the below for strange environments like Revo + Jupyter
            # where root handlers appear to already be set.
            # We don't want to spew to those consoles with DEBUG emissions
            n_logger.propagate = False

        module_logger.info(
            "\n\n********** STARTING CAPTURE FOR [%s] **********\n\n",
            self._context_name,
        )
        self._capturing = True

    def stop_capture(self) -> None:
        """Stop the capture of debug logs."""
        if not self._capturing:
            module_logger.warning("Debug logs are already disabled.")
            return

        module_logger.info(
            "\n\n********** STOPPING CAPTURE FOR [%s] **********\n\n",
            self._context_name,
        )
        print("Disabling log capture. Resulting file is at {}".format(self._filename))

        for namespace in self._namespaces:
            module_logger.debug("Removing [%s] debug logs to this file", namespace)
            n_logger = logging.getLogger(namespace)
            n_logger.removeHandler(self._handler)

        self._capturing = False

    def __enter__(self) -> None:
        self.start_capture()

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        self.stop_capture()


_debugging_enabled = False


def debug_sdk() -> None:
    global _debugging_enabled  # pylint: disable=global-statement
    if _debugging_enabled:
        module_logger.warning("Debug logs are already enabled at %s", LOG_FILE)
        return

    formatter = logging.Formatter(LOG_FORMAT)
    formatter.converter = time.gmtime

    file_handler = logging.FileHandler(filename=LOG_FILE, encoding="utf-8")
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(formatter)

    module_logger.info("Debug logs are being sent to %s", LOG_FILE)

    for namespace in INTERESTING_NAMESPACES:
        module_logger.debug("Adding [%s] debug logs to this file", namespace)
        n_logger = logging.getLogger(namespace)
        n_logger.setLevel(logging.DEBUG)
        n_logger.addHandler(file_handler)
        # We do the below for strange environments like Revo + Jupyter
        # where root handlers appear to already be set.
        # We don't want to spew to those consoles with DEBUG emissions
        n_logger.propagate = 0  # type: ignore[assignment]

    _debugging_enabled = True



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=unused-argument

from marshmallow import fields, post_load

from azure.ai.ml._schema.core.schema_meta import PatchedSchemaMeta


class ResourceConfigurationSchema(metaclass=PatchedSchemaMeta):
    instance_count = fields.Int()
    instance_type = fields.Str(metadata={"description": "The instance type to make available to this job."})
    properties = fields.Dict(keys=fields.Str())

    @post_load
    def make(self, data, **kwargs):
        from azure.ai.ml.entities import ResourceConfiguration

        return ResourceConfiguration(**data)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=unused-argument

import logging
from typing import Any

from marshmallow import fields, post_load

from azure.ai.ml._schema.core.schema import PatchedSchemaMeta

module_logger = logging.getLogger(__name__)


class LivenessProbeSchema(metaclass=PatchedSchemaMeta):
    period = fields.Int()
    initial_delay = fields.Int()
    timeout = fields.Int()
    success_threshold = fields.Int()
    failure_threshold = fields.Int()

    @post_load
    def make(self, data: Any, **kwargs: Any) -> Any:
        from azure.ai.ml.entities import ProbeSettings

        return ProbeSettings(**data)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
# pylint: disable=protected-access,too-many-lines
import copy
import logging
import os
from enum import Enum
from os import PathLike
from typing import Any, Dict, List, Optional, Tuple, Union, cast, overload

from marshmallow import INCLUDE, Schema

from azure.ai.ml._restclient.v2025_01_01_preview.models import CommandJob as RestCommandJob
from azure.ai.ml._restclient.v2025_01_01_preview.models import JobBase
from azure.ai.ml._schema.core.fields import ExperimentalField, NestedField, UnionField
from azure.ai.ml._schema.job.command_job import CommandJobSchema
from azure.ai.ml._schema.job.identity import AMLTokenIdentitySchema, ManagedIdentitySchema, UserIdentitySchema
from azure.ai.ml._schema.job.services import JobServiceSchema
from azure.ai.ml.constants._common import BASE_PATH_CONTEXT_KEY, LOCAL_COMPUTE_PROPERTY, LOCAL_COMPUTE_TARGET
from azure.ai.ml.constants._component import ComponentSource, NodeType
from azure.ai.ml.entities._assets import Environment
from azure.ai.ml.entities._component.command_component import CommandComponent
from azure.ai.ml.entities._credentials import (
    AmlTokenConfiguration,
    ManagedIdentityConfiguration,
    UserIdentityConfiguration,
    _BaseJobIdentityConfiguration,
)
from azure.ai.ml.entities._inputs_outputs import Input, Output
from azure.ai.ml.entities._job._input_output_helpers import from_rest_data_outputs, from_rest_inputs_to_dataset_literal
from azure.ai.ml.entities._job.command_job import CommandJob
from azure.ai.ml.entities._job.distribution import (
    DistributionConfiguration,
    MpiDistribution,
    PyTorchDistribution,
    RayDistribution,
    TensorFlowDistribution,
)
from azure.ai.ml.entities._job.job_limits import CommandJobLimits
from azure.ai.ml.entities._job.job_resource_configuration import JobResourceConfiguration
from azure.ai.ml.entities._job.job_service import (
    JobService,
    JobServiceBase,
    JupyterLabJobService,
    SshJobService,
    TensorBoardJobService,
    VsCodeJobService,
)
from azure.ai.ml.entities._job.queue_settings import QueueSettings
from azure.ai.ml.entities._job.sweep.early_termination_policy import EarlyTerminationPolicy
from azure.ai.ml.entities._job.sweep.objective import Objective
from azure.ai.ml.entities._job.sweep.search_space import (
    Choice,
    LogNormal,
    LogUniform,
    Normal,
    QLogNormal,
    QLogUniform,
    QNormal,
    QUniform,
    Randint,
    SweepDistribution,
    Uniform,
)
from azure.ai.ml.entities._system_data import SystemData
from azure.ai.ml.exceptions import ErrorCategory, ErrorTarget, ValidationErrorType, ValidationException

from ..._schema import PathAwareSchema
from ..._schema.job.distribution import (
    MPIDistributionSchema,
    PyTorchDistributionSchema,
    RayDistributionSchema,
    TensorFlowDistributionSchema,
)
from .._job.pipeline._io import NodeWithGroupInputMixin
from .._util import (
    convert_ordered_dict_to_dict,
    from_rest_dict_to_dummy_rest_object,
    get_rest_dict_for_node_attrs,
    load_from_dict,
    validate_attribute_type,
)
from .base_node import BaseNode
from .sweep import Sweep

module_logger = logging.getLogger(__name__)


class Command(BaseNode, NodeWithGroupInputMixin):
    """Base class for command node, used for command component version consumption.

    You should not instantiate this class directly. Instead, you should create it using the builder function: command().

    :keyword component: The ID or instance of the command component or job to be run for the step.
    :paramtype component: Union[str, ~azure.ai.ml.entities.CommandComponent]
    :keyword compute: The compute target the job will run on.
    :paramtype compute: Optional[str]
    :keyword inputs: A mapping of input names to input data sources used in the job.
    :paramtype inputs: Optional[dict[str, Union[
        ~azure.ai.ml.Input, str, bool, int, float, Enum]]]
    :keyword outputs: A mapping of output names to output data sources used in the job.
    :paramtype outputs: Optional[dict[str, Union[str, ~azure.ai.ml.Output]]]
    :keyword limits: The limits for the command component or job.
    :paramtype limits: ~azure.ai.ml.entities.CommandJobLimits
    :keyword identity: The identity that the command job will use while running on compute.
    :paramtype identity: Optional[Union[
        dict[str, str],
        ~azure.ai.ml.entities.ManagedIdentityConfiguration,
        ~azure.ai.ml.entities.AmlTokenConfiguration,
        ~azure.ai.ml.entities.UserIdentityConfiguration]]
    :keyword distribution: The configuration for distributed jobs.
    :paramtype distribution: Optional[Union[dict, ~azure.ai.ml.PyTorchDistribution, ~azure.ai.ml.MpiDistribution,
        ~azure.ai.ml.TensorFlowDistribution, ~azure.ai.ml.RayDistribution]]
    :keyword environment: The environment that the job will run in.
    :paramtype environment: Optional[Union[str, ~azure.ai.ml.entities.Environment]]
    :keyword environment_variables:  A dictionary of environment variable names and values.
        These environment variables are set on the process where the user script is being executed.
    :paramtype environment_variables: Optional[dict[str, str]]
    :keyword resources: The compute resource configuration for the command.
    :paramtype resources: Optional[~azure.ai.ml.entities.JobResourceConfiguration]
    :keyword services: The interactive services for the node. This is an experimental parameter, and may change at any
        time. Please see https://aka.ms/azuremlexperimental for more information.
    :paramtype services: Optional[dict[str, Union[~azure.ai.ml.entities.JobService,
        ~azure.ai.ml.entities.JupyterLabJobService,
        ~azure.ai.ml.entities.SshJobService, ~azure.ai.ml.entities.TensorBoardJobService,
        ~azure.ai.ml.entities.VsCodeJobService]]]
    :keyword queue_settings: Queue settings for the job.
    :paramtype queue_settings: Optional[~azure.ai.ml.entities.QueueSettings]
    :keyword parent_job_name: parent job id for command job
    :paramtype parent_job_name: Optional[str]
    :raises ~azure.ai.ml.exceptions.ValidationException: Raised if Command cannot be successfully validated.
        Details will be provided in the error message.
    """

    # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        *,
        component: Union[str, CommandComponent],
        compute: Optional[str] = None,
        inputs: Optional[
            Dict[
                str,
                Union[
                    Input,
                    str,
                    bool,
                    int,
                    float,
                    Enum,
                ],
            ]
        ] = None,
        outputs: Optional[Dict[str, Union[str, Output]]] = None,
        limits: Optional[CommandJobLimits] = None,
        identity: Optional[
            Union[Dict, ManagedIdentityConfiguration, AmlTokenConfiguration, UserIdentityConfiguration]
        ] = None,
        distribution: Optional[
            Union[
                Dict,
                MpiDistribution,
                TensorFlowDistribution,
                PyTorchDistribution,
                RayDistribution,
                DistributionConfiguration,
            ]
        ] = None,
        environment: Optional[Union[Environment, str]] = None,
        environment_variables: Optional[Dict] = None,
        resources: Optional[JobResourceConfiguration] = None,
        services: Optional[
            Dict[str, Union[JobService, JupyterLabJobService, SshJobService, TensorBoardJobService, VsCodeJobService]]
        ] = None,
        queue_settings: Optional[QueueSettings] = None,
        parent_job_name: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        # validate init params are valid type
        validate_attribute_type(attrs_to_check=locals(), attr_type_map=self._attr_type_map())

        # resolve normal dict to dict[str, JobService]
        services = _resolve_job_services(services)
        kwargs.pop("type", None)
        self._parameters: dict = kwargs.pop("parameters", {})
        BaseNode.__init__(
            self,
            type=NodeType.COMMAND,
            inputs=inputs,
            outputs=outputs,
            component=component,
            compute=compute,
            services=services,
            **kwargs,
        )

        # init mark for _AttrDict
        self._init = True
        # initialize command job properties
        self.limits = limits
        self.identity = identity
        self._distribution = distribution
        self.environment_variables = {} if environment_variables is None else environment_variables
        self.environment: Any = environment
        self._resources = resources
        self._services = services
        self.queue_settings = queue_settings
        self.parent_job_name = parent_job_name

        if isinstance(self.component, CommandComponent):
            self.resources = self.resources or self.component.resources  # type: ignore[assignment]
            self.distribution = self.distribution or self.component.distribution

        self._swept: bool = False
        self._init = False

    @classmethod
    def _get_supported_inputs_types(cls) -> Tuple:
        supported_types = super()._get_supported_inputs_types() or ()
        return (
            SweepDistribution,
            *supported_types,
        )

    @classmethod
    def _get_supported_outputs_types(cls) -> Tuple:
        return str, Output

    @property
    def parameters(self) -> Dict[str, str]:
        """MLFlow parameters to be logged during the job.

        :return: The MLFlow parameters to be logged during the job.
        :rtype: dict[str, str]
        """
        return self._parameters

    @property
    def distribution(
        self,
    ) -> Optional[
        Union[
            Dict,
            MpiDistribution,
            TensorFlowDistribution,
            PyTorchDistribution,
            RayDistribution,
            DistributionConfiguration,
        ]
    ]:
        """The configuration for the distributed command component or job.

        :return: The configuration for distributed jobs.
        :rtype: Union[~azure.ai.ml.PyTorchDistribution, ~azure.ai.ml.MpiDistribution,
            ~azure.ai.ml.TensorFlowDistribution, ~azure.ai.ml.RayDistribution]
        """
        return self._distribution

    @distribution.setter
    def distribution(
        self,
        value: Union[Dict, PyTorchDistribution, TensorFlowDistribution, MpiDistribution, RayDistribution],
    ) -> None:
        """Sets the configuration for the distributed command component or job.

        :param value: The configuration for distributed jobs.
        :type value: Union[dict, ~azure.ai.ml.PyTorchDistribution, ~azure.ai.ml.MpiDistribution,
            ~azure.ai.ml.TensorFlowDistribution, ~azure.ai.ml.RayDistribution]
        """
        if isinstance(value, dict):
            dist_schema = UnionField(
                [
                    NestedField(PyTorchDistributionSchema, unknown=INCLUDE),
                    NestedField(TensorFlowDistributionSchema, unknown=INCLUDE),
                    NestedField(MPIDistributionSchema, unknown=INCLUDE),
                    ExperimentalField(NestedField(RayDistributionSchema, unknown=INCLUDE)),
                ]
            )
            value = dist_schema._deserialize(value=value, attr=None, data=None)
        self._distribution = value

    @property
    def resources(self) -> JobResourceConfiguration:
        """The compute resource configuration for the command component or job.

        :rtype: ~azure.ai.ml.entities.JobResourceConfiguration
        """
        return cast(JobResourceConfiguration, self._resources)

    @resources.setter
    def resources(self, value: Union[Dict, JobResourceConfiguration]) -> None:
        """Sets the compute resource configuration for the command component or job.

        :param value: The compute resource configuration for the command component or job.
        :type value: Union[dict, ~azure.ai.ml.entities.JobResourceConfiguration]
        """
        if isinstance(value, dict):
            value = JobResourceConfiguration(**value)
        self._resources = value

    @property
    def queue_settings(self) -> Optional[QueueSettings]:
        """The queue settings for the command component or job.

        :return: The queue settings for the command component or job.
        :rtype: ~azure.ai.ml.entities.QueueSettings
        """
        return self._queue_settings

    @queue_settings.setter
    def queue_settings(self, value: Union[Dict, QueueSettings]) -> None:
        """Sets the queue settings for the command component or job.

        :param value: The queue settings for the command component or job.
        :type value: Union[dict, ~azure.ai.ml.entities.QueueSettings]
        """
        if isinstance(value, dict):
            value = QueueSettings(**value)
        self._queue_settings = value

    @property
    def identity(
        self,
    ) -> Optional[Union[Dict, ManagedIdentityConfiguration, AmlTokenConfiguration, UserIdentityConfiguration]]:
        """The identity that the job will use while running on compute.

        :return: The identity that the job will use while running on compute.
        :rtype: Optional[Union[~azure.ai.ml.ManagedIdentityConfiguration, ~azure.ai.ml.AmlTokenConfiguration,
            ~azure.ai.ml.UserIdentityConfiguration]]
        """
        return self._identity

    @identity.setter
    def identity(
        self,
        value: Optional[Union[Dict, ManagedIdentityConfiguration, AmlTokenConfiguration, UserIdentityConfiguration]],
    ) -> None:
        """Sets the identity that the job will use while running on compute.

        :param value: The identity that the job will use while running on compute.
        :type value: Union[dict[str, str], ~azure.ai.ml.ManagedIdentityConfiguration,
            ~azure.ai.ml.AmlTokenConfiguration, ~azure.ai.ml.UserIdentityConfiguration]
        """
        if isinstance(value, dict):
            identity_schema = UnionField(
                [
                    NestedField(ManagedIdentitySchema, unknown=INCLUDE),
                    NestedField(AMLTokenIdentitySchema, unknown=INCLUDE),
                    NestedField(UserIdentitySchema, unknown=INCLUDE),
                ]
            )
            value = identity_schema._deserialize(value=value, attr=None, data=None)
        self._identity = value

    @property
    def services(
        self,
    ) -> Optional[
        Dict[str, Union[JobService, JupyterLabJobService, SshJobService, TensorBoardJobService, VsCodeJobService]]
    ]:
        """The interactive services for the node.

        This is an experimental parameter, and may change at any time.
        Please see https://aka.ms/azuremlexperimental for more information.

        :rtype: dict[str, Union[~azure.ai.ml.entities.JobService, ~azure.ai.ml.entities.JupyterLabJobService,
            ~azure.ai.ml.entities.SshJobService, ~azure.ai.ml.entities.TensorBoardJobService,
            ~azure.ai.ml.entities.VsCodeJobService]]
        """
        return self._services

    @services.setter
    def services(
        self,
        value: Dict,
    ) -> None:
        """Sets the interactive services for the node.

        This is an experimental parameter, and may change at any time.
        Please see https://aka.ms/azuremlexperimental for more information.

        :param value: The interactive services for the node.
        :type value: dict[str, Union[~azure.ai.ml.entities.JobService, ~azure.ai.ml.entities.JupyterLabJobService,
            ~azure.ai.ml.entities.SshJobService, ~azure.ai.ml.entities.TensorBoardJobService,
            ~azure.ai.ml.entities.VsCodeJobService]]
        """
        self._services = _resolve_job_services(value)  # type: ignore[assignment]

    @property
    def component(self) -> Union[str, CommandComponent]:
        """The ID or instance of the command component or job to be run for the step.

        :return: The ID or instance of the command component or job to be run for the step.
        :rtype: Union[str, ~azure.ai.ml.entities.CommandComponent]
        """
        return self._component

    @property
    def command(self) -> Optional[str]:
        """The command to be executed.

        :rtype: Optional[str]
        """
        # the same as code
        if not isinstance(self.component, CommandComponent):
            return None

        if self.component.command is None:
            return None
        return str(self.component.command)

    @command.setter
    def command(self, value: str) -> None:
        """Sets the command to be executed.

        :param value: The command to be executed.
        :type value: str
        """
        if isinstance(self.component, CommandComponent):
            self.component.command = value
        else:
            msg = "Can't set command property for a registered component {}. Tried to set it to {}."
            raise ValidationException(
                message=msg.format(self.component, value),
                no_personal_data_message=msg,
                target=ErrorTarget.COMMAND_JOB,
                error_category=ErrorCategory.USER_ERROR,
                error_type=ValidationErrorType.INVALID_VALUE,
            )

    @property
    def code(self) -> Optional[Union[str, PathLike]]:
        """The source code to run the job.

        :rtype: Optional[Union[str, os.PathLike]]
        """
        # BaseNode is an _AttrDict to allow dynamic attributes, so that lower version of SDK can work with attributes
        # added in higher version of SDK.
        # self.code will be treated as an Arbitrary attribute if it raises AttributeError in getting
        # (when self.component doesn't have attribute code, self.component = 'azureml:xxx:1' e.g.
        # you may check _AttrDict._is_arbitrary_attr for detailed logic for Arbitrary judgement),
        # then its value will be set to _AttrDict and be deserialized as {"shape": {}} instead of None,
        # which is invalid in schema validation.
        if not isinstance(self.component, CommandComponent):
            return None

        if self.component.code is None:
            return None

        return str(self.component.code)

    @code.setter
    def code(self, value: str) -> None:
        """Sets the source code to run the job.

        :param value: The source code to run the job. Can be a local path or "http:", "https:", or "azureml:" url
            pointing to a remote location.
        :type value: str
        """
        if isinstance(self.component, CommandComponent):
            self.component.code = value
        else:
            msg = "Can't set code property for a registered component {}"
            raise ValidationException(
                message=msg.format(self.component),
                no_personal_data_message=msg,
                target=ErrorTarget.COMMAND_JOB,
                error_category=ErrorCategory.USER_ERROR,
                error_type=ValidationErrorType.INVALID_VALUE,
            )

    def set_resources(
        self,
        *,
        instance_type: Optional[Union[str, List[str]]] = None,
        instance_count: Optional[int] = None,
        locations: Optional[List[str]] = None,
        properties: Optional[Dict] = None,
        docker_args: Optional[Union[str, List[str]]] = None,
        shm_size: Optional[str] = None,
        # pylint: disable=unused-argument
        **kwargs: Any,
    ) -> None:
        """Set resources for Command.

        :keyword instance_type: The type of compute instance to run the job on. If not specified, the job will run on
            the default compute target.
        :paramtype instance_type: Optional[Union[str, List[str]]]
        :keyword instance_count: The number of instances to run the job on. If not specified, the job will run on a
            single instance.
        :paramtype instance_count: Optional[int]
        :keyword locations: The list of locations where the job will run. If not specified, the job will run on the
            default compute target.
        :paramtype locations: Optional[List[str]]
        :keyword properties: The properties of the job.
        :paramtype properties: Optional[dict]
        :keyword docker_args: The Docker arguments for the job.
        :paramtype docker_args: Optional[Union[str,List[str]]]
        :keyword shm_size: The size of the docker container's shared memory block. This should be in the
            format of (number)(unit) where the number has to be greater than 0 and the unit can be one of
            b(bytes), k(kilobytes), m(megabytes), or g(gigabytes).
        :paramtype shm_size: Optional[str]

        .. admonition:: Example:

            .. literalinclude:: ../samples/ml_samples_command_configurations.py
                :start-after: [START command_set_resources]
                :end-before: [END command_set_resources]
                :language: python
                :dedent: 8
                :caption: Setting resources on a Command.
        """
        if self.resources is None:
            self.resources = JobResourceConfiguration()

        if locations is not None:
            self.resources.locations = locations
        if instance_type is not None:
            self.resources.instance_type = instance_type
        if instance_count is not None:
            self.resources.instance_count = instance_count
        if properties is not None:
            self.resources.properties = properties
        if docker_args is not None:
            self.resources.docker_args = docker_args
        if shm_size is not None:
            self.resources.shm_size = shm_size

        # Save the resources to internal component as well, otherwise calling sweep() will loose the settings
        if isinstance(self.component, CommandComponent):
            self.component.resources = self.resources

    def set_limits(self, *, timeout: int, **kwargs: Any) -> None:  # pylint: disable=unused-argument
        """Set limits for Command.

        :keyword timeout: The timeout for the job in seconds.
        :paramtype timeout: int

        .. admonition:: Example:

            .. literalinclude:: ../samples/ml_samples_command_configurations.py
                :start-after: [START command_set_limits]
                :end-before: [END command_set_limits]
                :language: python
                :dedent: 8
                :caption: Setting a timeout limit of 10 seconds on a Command.
        """
        if isinstance(self.limits, CommandJobLimits):
            self.limits.timeout = timeout
        else:
            self.limits = CommandJobLimits(timeout=timeout)

    def set_queue_settings(self, *, job_tier: Optional[str] = None, priority: Optional[str] = None) -> None:
        """Set QueueSettings for the job.

        :keyword job_tier: The job tier. Accepted values are "Spot", "Basic", "Standard", or "Premium".
        :paramtype job_tier: Optional[str]
        :keyword priority:  The priority of the job on the compute. Defaults to "Medium".
        :paramtype priority: Optional[str]

        .. admonition:: Example:

            .. literalinclude:: ../samples/ml_samples_command_configurations.py
                :start-after: [START command_set_queue_settings]
                :end-before: [END command_set_queue_settings]
                :language: python
                :dedent: 8
                :caption: Configuring queue settings on a Command.
        """
        if isinstance(self.queue_settings, QueueSettings):
            self.queue_settings.job_tier = job_tier
            self.queue_settings.priority = priority
        else:
            self.queue_settings = QueueSettings(job_tier=job_tier, priority=priority)

    def sweep(
        self,
        *,
        primary_metric: str,
        goal: str,
        sampling_algorithm: str = "random",
        compute: Optional[str] = None,
        max_concurrent_trials: Optional[int] = None,
        max_total_trials: Optional[int] = None,
        timeout: Optional[int] = None,
        trial_timeout: Optional[int] = None,
        early_termination_policy: Optional[Union[EarlyTerminationPolicy, str]] = None,
        search_space: Optional[
            Dict[
                str,
                Union[
                    Choice, LogNormal, LogUniform, Normal, QLogNormal, QLogUniform, QNormal, QUniform, Randint, Uniform
                ],
            ]
        ] = None,
        identity: Optional[
            Union[ManagedIdentityConfiguration, AmlTokenConfiguration, UserIdentityConfiguration]
        ] = None,
        queue_settings: Optional[QueueSettings] = None,
        job_tier: Optional[str] = None,
        priority: Optional[str] = None,
    ) -> Sweep:
        """Turns the command into a sweep node with extra sweep run setting. The command component
        in the current command node will be used as its trial component. A command node can sweep
        multiple times, and the generated sweep node will share the same trial component.

        :keyword primary_metric: The primary metric of the sweep objective - e.g. AUC (Area Under the Curve).
            The metric must be logged while running the trial component.
        :paramtype primary_metric: str
        :keyword goal: The goal of the Sweep objective. Accepted values are "minimize" or "maximize".
        :paramtype goal: str
        :keyword sampling_algorithm: The sampling algorithm to use inside the search space.
            Acceptable values are "random", "grid", or "bayesian". Defaults to "random".
        :paramtype sampling_algorithm: str
        :keyword compute: The target compute to run the node on. If not specified, the current node's compute
            will be used.
        :paramtype compute: Optional[str]
        :keyword max_total_trials: The maximum number of total trials to run. This value will overwrite the value in
            CommandJob.limits if specified.
        :paramtype max_total_trials: Optional[int]
        :keyword max_concurrent_trials: The maximum number of concurrent trials for the Sweep job.
        :paramtype max_concurrent_trials: Optional[int]
        :keyword timeout: The maximum run duration in seconds, after which the job will be cancelled.
        :paramtype timeout: Optional[int]
        :keyword trial_timeout: The Sweep Job trial timeout value, in seconds.
        :paramtype trial_timeout: Optional[int]
        :keyword early_termination_policy: The early termination policy of the sweep node. Acceptable
            values are "bandit", "median_stopping", or "truncation_selection". Defaults to None.
        :paramtype early_termination_policy: Optional[Union[~azure.ai.ml.sweep.BanditPolicy,
            ~azure.ai.ml.sweep.TruncationSelectionPolicy, ~azure.ai.ml.sweep.MedianStoppingPolicy, str]]
        :keyword identity: The identity that the job will use while running on compute.
        :paramtype identity: Optional[Union[
            ~azure.ai.ml.ManagedIdentityConfiguration,
            ~azure.ai.ml.AmlTokenConfiguration,
            ~azure.ai.ml.UserIdentityConfiguration]]
        :keyword search_space: The search space to use for the sweep job.
        :paramtype search_space: Optional[Dict[str, Union[
            Choice,
            LogNormal,
            LogUniform,
            Normal,
            QLogNormal,
            QLogUniform,
            QNormal,
            QUniform,
            Randint,
            Uniform

        ]]]

        :keyword queue_settings: The queue settings for the job.
        :paramtype queue_settings: Optional[~azure.ai.ml.entities.QueueSettings]
        :keyword job_tier: **Experimental** The job tier. Accepted values are "Spot", "Basic",
            "Standard", or "Premium".
        :paramtype job_tier: Optional[str]
        :keyword priority: **Experimental** The compute priority. Accepted values are "low",
            "medium", and "high".
        :paramtype priority: Optional[str]
        :return: A Sweep node with the component from current Command node as its trial component.
        :rtype: ~azure.ai.ml.entities.Sweep

        .. admonition:: Example:

            .. literalinclude:: ../samples/ml_samples_sweep_configurations.py
                :start-after: [START configure_sweep_job_bandit_policy]
                :end-before: [END configure_sweep_job_bandit_policy]
                :language: python
                :dedent: 8
                :caption: Creating a Sweep node from a Command job.
        """
        self._swept = True
        # inputs & outputs are already built in source Command obj
        inputs, inputs_search_space = Sweep._get_origin_inputs_and_search_space(self.inputs)
        if search_space:
            inputs_search_space.update(search_space)

        if not queue_settings:
            queue_settings = self.queue_settings
        if queue_settings is not None:
            if job_tier is not None:
                queue_settings.job_tier = job_tier
            if priority is not None:
                queue_settings.priority = priority

        sweep_node = Sweep(
            trial=copy.deepcopy(
                self.component
            ),  # Make a copy of the underneath Component so that the original node can still be used.
            compute=self.compute if compute is None else compute,
            objective=Objective(goal=goal, primary_metric=primary_metric),
            sampling_algorithm=sampling_algorithm,
            inputs=inputs,
            outputs=self._get_origin_job_outputs(),
            search_space=inputs_search_space,
            early_termination=early_termination_policy,
            name=self.name,
            description=self.description,
            display_name=self.display_name,
            tags=self.tags,
            properties=self.properties,
            experiment_name=self.experiment_name,
            identity=self.identity if not identity else identity,
            _from_component_func=True,
            queue_settings=queue_settings,
        )
        sweep_node.set_limits(
            max_total_trials=max_total_trials,
            max_concurrent_trials=max_concurrent_trials,
            timeout=timeout,
            trial_timeout=trial_timeout,
        )
        return sweep_node

    @classmethod
    def _attr_type_map(cls) -> dict:
        return {
            "component": (str, CommandComponent),
            "environment": (str, Environment),
            "environment_variables": dict,
            "resources": (dict, JobResourceConfiguration),
            "limits": (dict, CommandJobLimits),
            "code": (str, os.PathLike),
        }

    def _to_job(self) -> CommandJob:
        if isinstance(self.component, CommandComponent):
            return CommandJob(
                id=self.id,
                name=self.name,
                display_name=self.display_name,
                description=self.description,
                tags=self.tags,
                properties=self.properties,
                command=self.component.command,
                experiment_name=self.experiment_name,
                code=self.component.code,
                compute=self.compute,
                status=self.status,
                environment=self.environment,
                distribution=self.distribution,
                identity=self.identity,
                environment_variables=self.environment_variables,
                resources=self.resources,
                limits=self.limits,
                inputs=self._job_inputs,
                outputs=self._job_outputs,
                services=self.services,
                creation_context=self.creation_context,
                parameters=self.parameters,
                queue_settings=self.queue_settings,
                parent_job_name=self.parent_job_name,
            )

        return CommandJob(
            id=self.id,
            name=self.name,
            display_name=self.display_name,
            description=self.description,
            tags=self.tags,
            properties=self.properties,
            command=None,
            experiment_name=self.experiment_name,
            code=None,
            compute=self.compute,
            status=self.status,
            environment=self.environment,
            distribution=self.distribution,
            identity=self.identity,
            environment_variables=self.environment_variables,
            resources=self.resources,
            limits=self.limits,
            inputs=self._job_inputs,
            outputs=self._job_outputs,
            services=self.services,
            creation_context=self.creation_context,
            parameters=self.parameters,
            queue_settings=self.queue_settings,
            parent_job_name=self.parent_job_name,
        )

    @classmethod
    def _picked_fields_from_dict_to_rest_object(cls) -> List[str]:
        return ["resources", "distribution", "limits", "environment_variables", "queue_settings"]

    def _to_rest_object(self, **kwargs: Any) -> dict:
        rest_obj = super()._to_rest_object(**kwargs)
        for key, value in {
            "componentId": self._get_component_id(),
            "distribution": get_rest_dict_for_node_attrs(self.distribution, clear_empty_value=True),
            "limits": get_rest_dict_for_node_attrs(self.limits, clear_empty_value=True),
            "resources": get_rest_dict_for_node_attrs(self.resources, clear_empty_value=True),
            "services": get_rest_dict_for_node_attrs(self.services),
            "identity": get_rest_dict_for_node_attrs(self.identity),
            "queue_settings": get_rest_dict_for_node_attrs(self.queue_settings, clear_empty_value=True),
        }.items():
            if value is not None:
                rest_obj[key] = value
        return cast(dict, convert_ordered_dict_to_dict(rest_obj))

    @classmethod
    def _load_from_dict(cls, data: Dict, context: Dict, additional_message: str, **kwargs: Any) -> "Command":
        from .command_func import command

        loaded_data = load_from_dict(CommandJobSchema, data, context, additional_message, **kwargs)

        # resources a limits properties are flatten in command() function, exact them and set separately
        resources = loaded_data.pop("resources", None)
        limits = loaded_data.pop("limits", None)

        command_job: Command = command(base_path=context[BASE_PATH_CONTEXT_KEY], **loaded_data)

        command_job.resources = resources
        command_job.limits = limits
        return command_job

    @classmethod
    def _from_rest_object_to_init_params(cls, obj: dict) -> Dict:
        obj = BaseNode._from_rest_object_to_init_params(obj)

        if "resources" in obj and obj["resources"]:
            obj["resources"] = JobResourceConfiguration._from_rest_object(obj["resources"])

        # services, sweep won't have services
        if "services" in obj and obj["services"]:
            # pipeline node rest object are dicts while _from_rest_job_services expect RestJobService
            services = {}
            for service_name, service in obj["services"].items():
                # in rest object of a pipeline job, service will be transferred to a dict as
                # it's attributes of a node, but JobService._from_rest_object expect a
                # RestJobService, so we need to convert it back. Here we convert the dict to a
                # dummy rest object which may work as a RestJobService instead.
                services[service_name] = from_rest_dict_to_dummy_rest_object(service)
            obj["services"] = JobServiceBase._from_rest_job_services(services)

        # handle limits
        if "limits" in obj and obj["limits"]:
            obj["limits"] = CommandJobLimits._from_rest_object(obj["limits"])

        if "identity" in obj and obj["identity"]:
            obj["identity"] = _BaseJobIdentityConfiguration._from_rest_object(obj["identity"])

        if "queue_settings" in obj and obj["queue_settings"]:
            obj["queue_settings"] = QueueSettings._from_rest_object(obj["queue_settings"])

        return obj

    @classmethod
    def _load_from_rest_job(cls, obj: JobBase) -> "Command":
        from .command_func import command

        rest_command_job: RestCommandJob = obj.properties

        command_job: Command = command(
            name=obj.name,
            display_name=rest_command_job.display_name,
            description=rest_command_job.description,
            tags=rest_command_job.tags,
            properties=rest_command_job.properties,
            command=rest_command_job.command,
            experiment_name=rest_command_job.experiment_name,
            services=JobServiceBase._from_rest_job_services(rest_command_job.services),
            status=rest_command_job.status,
            creation_context=SystemData._from_rest_object(obj.system_data) if obj.system_data else None,
            code=rest_command_job.code_id,
            compute=rest_command_job.compute_id,
            environment=rest_command_job.environment_id,
            distribution=DistributionConfiguration._from_rest_object(rest_command_job.distribution),
            parameters=rest_command_job.parameters,
            identity=(
                _BaseJobIdentityConfiguration._from_rest_object(rest_command_job.identity)
                if rest_command_job.identity
                else None
            ),
            environment_variables=rest_command_job.environment_variables,
            inputs=from_rest_inputs_to_dataset_literal(rest_command_job.inputs),
            outputs=from_rest_data_outputs(rest_command_job.outputs),
        )
        command_job._id = obj.id
        command_job.resources = cast(
            JobResourceConfiguration, JobResourceConfiguration._from_rest_object(rest_command_job.resources)
        )
        command_job.limits = CommandJobLimits._from_rest_object(rest_command_job.limits)
        command_job.queue_settings = QueueSettings._from_rest_object(rest_command_job.queue_settings)
        if isinstance(command_job.component, CommandComponent):
            command_job.component._source = (
                ComponentSource.REMOTE_WORKSPACE_JOB
            )  # This is used by pipeline job telemetries.

        # Handle special case of local job
        if (
            command_job.resources is not None
            and command_job.resources.properties is not None
            and command_job.resources.properties.get(LOCAL_COMPUTE_PROPERTY, None)
        ):
            command_job.compute = LOCAL_COMPUTE_TARGET
            command_job.resources.properties.pop(LOCAL_COMPUTE_PROPERTY)
        return command_job

    def _build_inputs(self) -> Dict:
        inputs = super(Command, self)._build_inputs()
        built_inputs = {}
        # Validate and remove non-specified inputs
        for key, value in inputs.items():
            if value is not None:
                built_inputs[key] = value

        return built_inputs

    @classmethod
    def _create_schema_for_validation(cls, context: Any) -> Union[PathAwareSchema, Schema]:
        from azure.ai.ml._schema.pipeline import CommandSchema

        return CommandSchema(context=context)

    # pylint: disable-next=docstring-missing-param
    def __call__(self, *args: Any, **kwargs: Any) -> "Command":
        """Call Command as a function will return a new instance each time.

        :return: A Command node
        :rtype: Command
        """
        if isinstance(self._component, CommandComponent):
            # call this to validate inputs
            node: Command = self._component(*args, **kwargs)
            # merge inputs
            for name, original_input in self.inputs.items():
                if name not in kwargs:
                    # use setattr here to make sure owner of input won't change
                    setattr(node.inputs, name, original_input._data)
                    node._job_inputs[name] = original_input._data
                # get outputs
            for name, original_output in self.outputs.items():
                # use setattr here to make sure owner of input won't change
                if not isinstance(original_output, str):
                    setattr(node.outputs, name, original_output._data)
                    node._job_outputs[name] = original_output._data
            self._refine_optional_inputs_with_no_value(node, kwargs)
            # set default values: compute, environment_variables, outputs
            # won't copy name to be able to distinguish if a node's name is assigned by user
            # e.g. node_1 = command_func()
            # In above example, node_1.name will be None so we can apply node_1 as it's name
            node.compute = self.compute
            node.tags = self.tags
            # Pass through the display name only if the display name is not system generated.
            node.display_name = self.display_name if self.display_name != self.name else None
            node.environment = copy.deepcopy(self.environment)
            # deep copy for complex object
            node.environment_variables = copy.deepcopy(self.environment_variables)
            node.limits = copy.deepcopy(self.limits)
            node.distribution = copy.deepcopy(self.distribution)
            node.resources = copy.deepcopy(self.resources)
            node.queue_settings = copy.deepcopy(self.queue_settings)
            node.services = copy.deepcopy(self.services)
            node.identity = copy.deepcopy(self.identity)
            return node
        msg = "Command can be called as a function only when referenced component is {}, currently got {}."
        raise ValidationException(
            message=msg.format(type(CommandComponent), self._component),
            no_personal_data_message=msg.format(type(CommandComponent), "self._component"),
            target=ErrorTarget.COMMAND_JOB,
            error_type=ValidationErrorType.INVALID_VALUE,
        )


@overload
def _resolve_job_services(services: Optional[Dict]): ...


@overload
def _resolve_job_services(
    services: Dict[str, Union[JobServiceBase, Dict]],
) -> Dict[str, Union[JobService, JupyterLabJobService, SshJobService, TensorBoardJobService, VsCodeJobService]]: ...


def _resolve_job_services(
    services: Optional[Dict[str, Union[JobServiceBase, Dict]]],
) -> Optional[Dict]:
    """Resolve normal dict to dict[str, JobService]

    :param services: A dict that maps service names to either a JobServiceBase object, or a Dict used to build one
    :type services: Optional[Dict[str, Union[JobServiceBase, Dict]]]
    :return:
        * None if `services` is None
        * A map of job service names to job services
    :rtype: Optional[
            Dict[str, Union[JobService, JupyterLabJobService, SshJobService, TensorBoardJobService, VsCodeJobService]]
        ]
    """
    if services is None:
        return None

    if not isinstance(services, dict):
        msg = f"Services must be a dict, got {type(services)} instead."
        raise ValidationException(
            message=msg,
            no_personal_data_message=msg,
            target=ErrorTarget.COMMAND_JOB,
            error_category=ErrorCategory.USER_ERROR,
        )

    result = {}
    for name, service in services.items():
        if isinstance(service, dict):
            service = load_from_dict(JobServiceSchema, service, context={BASE_PATH_CONTEXT_KEY: "."})
        elif not isinstance(
            service, (JobService, JupyterLabJobService, SshJobService, TensorBoardJobService, VsCodeJobService)
        ):
            msg = f"Service value for key {name!r} must be a dict or JobService object, got {type(service)} instead."
            raise ValidationException(
                message=msg,
                no_personal_data_message=msg,
                target=ErrorTarget.COMMAND_JOB,
                error_category=ErrorCategory.USER_ERROR,
            )
        result[name] = service
    return result



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=protected-access

import logging
from abc import ABC
from typing import Any, Dict, Generic, List, Optional, TypeVar

K = TypeVar("K")
V = TypeVar("V")


class _AttrDict(Generic[K, V], Dict, ABC):
    """This class is used for accessing values with instance.some_key. It supports the following scenarios:

    1. Setting arbitrary attribute, eg: obj.resource_layout.node_count = 2
      1.1 Setting same nested filed twice will return same object, eg:
              obj.resource_layout.node_count = 2
              obj.resource_layout.process_count_per_node = 2
              obj.resource_layout will be {"node_count": 2, "process_count_per_node": 2}
      1.2 Only public attribute is supported, eg: obj._resource_layout._node_count = 2 will raise AttributeError
      1.3 All set attribute can be recorded, eg:
              obj.target = "aml"
              obj.resource_layout.process_count_per_node = 2
              obj.get_attr() will return {"target": "aml", "resource_layout": {"process_count_per_node": 2}}
    2. Getting arbitrary attribute, getting non-exist attribute will return an empty dict.
    3. Calling arbitrary methods is not allowed, eg: obj.resource_layout() should raise AttributeError
    """

    def __init__(self, allowed_keys: Optional[Dict] = None, **kwargs: Any):
        """Initialize a attribute dictionary.

        :param allowed_keys: A dictionary of keys that allowed to set as arbitrary attributes. None means all keys can
            be set as arbitrary attributes.

        :type dict
        :param kwargs: A dictionary of additional configuration parameters.
        :type kwargs: dict
        """
        super(_AttrDict, self).__init__(**kwargs)
        if allowed_keys is None:
            # None allowed_keys means no restriction on keys can be set for _AttrDict
            self._allowed_keys = {}
            self._key_restriction = False
        else:
            # Otherwise use allowed_keys to restrict keys can be set for _AttrDict
            self._allowed_keys = dict(allowed_keys)
            self._key_restriction = True
        self._logger = logging.getLogger("attr_dict")

    def _initializing(self) -> bool:
        # use this to indicate ongoing init process, sub class need to make sure this return True during init process.
        return False

    def _get_attrs(self) -> dict:
        """Get all arbitrary attributes which has been set, empty values are excluded.

        :return: A dict which contains all arbitrary attributes set by user.
        :rtype: dict
        """

        # TODO: check this
        def remove_empty_values(data: Dict) -> Dict:
            if not isinstance(data, dict):
                return data
            # skip empty dicts as default value of _AttrDict is empty dict
            return {k: remove_empty_values(v) for k, v in data.items() if v or not isinstance(v, dict)}

        return remove_empty_values(self)

    def _is_arbitrary_attr(self, attr_name: str) -> bool:
        """Checks if a given attribute name should be treat as arbitrary attribute.

        Attributes inside _AttrDict can be non-arbitrary attribute or arbitrary attribute.
        Non-arbitrary attributes are normal attributes like other object which stores in self.__dict__.
        Arbitrary attributes are attributes stored in the dictionary it self, what makes it special it it's value
        can be an instance of _AttrDict
        Take `obj = _AttrDict(allowed_keys={"resource_layout": {"node_count": None}})` as an example.
        `obj.some_key` is accessing non-arbitrary attribute.
        `obj.resource_layout` is accessing arbitrary attribute, user can use `obj.resource_layout.node_count = 1` to
        assign value to it.

        :param attr_name: Attribute name
        :type attr_name: str
        :return: If the given attribute name should be treated as arbitrary attribute.
        :rtype: bool
        """
        # Internal attribute won't be set as arbitrary attribute.
        if attr_name.startswith("_"):
            return False
        # All attributes set in __init__ won't be set as arbitrary attribute
        if self._initializing():
            return False
        # If there's key restriction, only keys in it can be set as arbitrary attribute.
        if self._key_restriction and attr_name not in self._allowed_keys:
            return False
        # Attributes already in attribute dict will not be set as arbitrary attribute.
        try:
            self.__getattribute__(attr_name)
        except AttributeError:
            return True
        return False

    def __getattr__(self, key: Any) -> Any:
        if not self._is_arbitrary_attr(key):
            return super().__getattribute__(key)
        self._logger.debug("getting %s", key)
        try:
            return super().__getitem__(key)
        except KeyError:
            allowed_keys = self._allowed_keys.get(key, None) if self._key_restriction else None
            result: Any = _AttrDict(allowed_keys=allowed_keys)
            self.__setattr__(key, result)
            return result

    def __setattr__(self, key: Any, value: V) -> None:
        if not self._is_arbitrary_attr(key):
            super().__setattr__(key, value)
        else:
            self._logger.debug("setting %s to %s", key, value)
            super().__setitem__(key, value)

    def __setitem__(self, key: Any, value: V) -> None:
        self.__setattr__(key, value)

    def __getitem__(self, item: V) -> Any:
        # support attr_dict[item] since dumping it in marshmallow requires this.
        return self.__getattr__(item)

    def __dir__(self) -> List:
        # For Jupyter Notebook auto-completion
        return list(super().__dir__()) + list(self.keys())


def has_attr_safe(obj: Any, attr: Any) -> bool:
    if isinstance(obj, _AttrDict):
        has_attr = not obj._is_arbitrary_attr(attr)
    elif isinstance(obj, dict):
        return attr in obj
    else:
        has_attr = hasattr(obj, attr)
    return has_attr


def try_get_non_arbitrary_attr(obj: Any, attr: str) -> Optional[Any]:
    """Try to get non-arbitrary attribute for potential attribute dict.

    Will not create target attribute if it is an arbitrary attribute in _AttrDict.

    :param obj: The obj
    :type obj: Any
    :param attr: The attribute name
    :type attr: str
    :return: obj.attr
    :rtype: Any
    """
    if has_attr_safe(obj, attr):
        return obj[attr] if isinstance(obj, dict) else getattr(obj, attr)
    return None



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

import logging
import re
from typing import Any, Optional, Tuple, Union

from azure.ai.ml._scope_dependent_operations import OperationScope
from azure.ai.ml.constants._common import (
    ARM_ID_PREFIX,
    ASSET_ID_URI_REGEX_FORMAT,
    AZUREML_RESOURCE_PROVIDER,
    DATA_ARM_TYPE,
    DATASTORE_RESOURCE_ID,
    DATASTORE_SHORT_URI,
    LEVEL_ONE_NAMED_RESOURCE_ID_FORMAT,
    NAMED_RESOURCE_ID_FORMAT,
    NAMED_RESOURCE_ID_FORMAT_WITH_PARENT,
    PROVIDER_RESOURCE_ID_WITH_VERSION,
    REGISTRY_URI_REGEX_FORMAT,
    REGISTRY_VERSION_PATTERN,
    SINGULARITY_FULL_NAME_REGEX_FORMAT,
    SINGULARITY_ID_REGEX_FORMAT,
    SINGULARITY_SHORT_NAME_REGEX_FORMAT,
)
from azure.ai.ml.exceptions import ErrorCategory, ErrorTarget, ValidationErrorType, ValidationException

module_logger = logging.getLogger(__name__)


class AMLVersionedArmId(object):
    """Parser for versioned arm id: e.g. /subscription/.../code/my- code/versions/1.

    :param arm_id: The versioned ARM id.
    :type arm_id: str
    :raises ~azure.ai.ml.exceptions.ValidationException: Raised if the ARM id is incorrectly formatted.
    """

    REGEX_PATTERN = (
        "^/?subscriptions/([^/]+)/resourceGroups/(["
        "^/]+)/providers/Microsoft.MachineLearningServices/workspaces/([^/]+)/([^/]+)/([^/]+)/versions/(["
        "^/]+)"
    )

    def __init__(self, arm_id=None):
        self.is_registry_id = None
        if arm_id:
            match = re.match(AMLVersionedArmId.REGEX_PATTERN, arm_id)
            if match:
                self.subscription_id = match.group(1)
                self.resource_group_name = match.group(2)
                self.workspace_name = match.group(3)
                self.asset_type = match.group(4)
                self.asset_name = match.group(5)
                self.asset_version = match.group(6)
            else:
                match = re.match(REGISTRY_VERSION_PATTERN, arm_id)
                if match:
                    self.asset_name = match.group(3)
                    self.asset_version = match.group(4)
                    self.is_registry_id = True
                else:
                    msg = "Invalid AzureML ARM versioned Id {}"
                    raise ValidationException(
                        message=msg.format(arm_id),
                        no_personal_data_message=msg.format("[arm_id]"),
                        error_type=ValidationErrorType.INVALID_VALUE,
                        error_category=ErrorCategory.USER_ERROR,
                        target=ErrorTarget.ARM_RESOURCE,
                    )


def get_datastore_arm_id(
    datastore_name: Optional[str] = None, operation_scope: Optional[OperationScope] = None
) -> Optional[str]:
    return (
        DATASTORE_RESOURCE_ID.format(
            operation_scope.subscription_id,
            operation_scope.resource_group_name,
            operation_scope.workspace_name,
            datastore_name,
        )
        if datastore_name
        else None
    )


class AMLLabelledArmId(object):
    """Parser for versioned arm id: e.g. /subscription/.../code/my- code/labels/default.

    :param arm_id: The labelled ARM id.
    :type arm_id: str
    :raises ~azure.ai.ml.exceptions.ValidationException: Raised if the ARM id is incorrectly formatted.
    """

    REGEX_PATTERN = (
        "^/?subscriptions/([^/]+)/resourceGroups/(["
        "^/]+)/providers/Microsoft.MachineLearningServices/workspaces/([^/]+)/([^/]+)/([^/]+)/labels/(["
        "^/]+)"
    )

    def __init__(self, arm_id=None):
        self.is_registry_id = None
        if arm_id:
            match = re.match(AMLLabelledArmId.REGEX_PATTERN, arm_id)
            if match:
                self.subscription_id = match.group(1)
                self.resource_group_name = match.group(2)
                self.workspace_name = match.group(3)
                self.asset_type = match.group(4)
                self.asset_name = match.group(5)
                self.asset_label = match.group(6)
            else:
                match = re.match(REGISTRY_VERSION_PATTERN, arm_id)
                if match:
                    self.asset_name = match.group(3)
                    self.asset_label = match.group(4)
                    self.is_registry_id = True
                else:
                    msg = "Invalid AzureML ARM versioned Id {}"
                    raise ValidationException(
                        message=msg.format(arm_id),
                        no_personal_data_message=msg.format("[arm_id]"),
                        error_type=ValidationErrorType.INVALID_VALUE,
                        error_category=ErrorCategory.USER_ERROR,
                        target=ErrorTarget.ARM_RESOURCE,
                    )


class AMLNamedArmId:
    """Parser for named arm id (no version): e.g.

    /subscription/.../compute/cpu-cluster.

    :param arm_id: The named ARM id.
    :type arm_id: str
    :raises ~azure.ai.ml.exceptions.ValidationException~: Raised if the ARM id is incorrectly formatted.
    """

    REGEX_PATTERN = (
        "^/?subscriptions/([^/]+)/resourceGroups/(["
        "^/]+)/providers/Microsoft.MachineLearningServices/workspaces/([^/]+)/([^/]+)/([^/]+)"
    )

    REGEX_PATTERN_WITH_PARENT = (
        "^/?subscriptions/([^/]+)/resourceGroups/(["
        "^/]+)/providers/Microsoft.MachineLearningServices/workspaces/([^/]+)/([^/]+)/([^/]+)/([^/]+)/([^/]+)"
    )

    def __init__(self, arm_id=None):
        if arm_id:
            match = re.match(AMLNamedArmId.REGEX_PATTERN_WITH_PARENT, arm_id)
            if match:
                self.asset_name = match.group(7)
                self.asset_type = match.group(6)
                self.parent_asset_name = match.group(5)
                self.parent_azureml_type = match.group(4)
            if match is None:
                match = re.match(AMLNamedArmId.REGEX_PATTERN, arm_id)
                if match:
                    self.asset_type = match.group(4)
                    self.asset_name = match.group(5)
                    self.parent_asset_name = None
                    self.parent_azureml_type = None
                else:
                    msg = "Invalid AzureML ARM named Id {}"
                    raise ValidationException(
                        message=msg.format(arm_id),
                        no_personal_data_message=msg.format("[arm_id]"),
                        error_type=ValidationErrorType.INVALID_VALUE,
                        error_category=ErrorCategory.USER_ERROR,
                        target=ErrorTarget.ARM_RESOURCE,
                    )

            self.subscription_id = match.group(1)
            self.resource_group_name = match.group(2)
            self.workspace_name = match.group(3)


class AMLAssetId:
    REGEX_PATTERN = ASSET_ID_URI_REGEX_FORMAT

    def __init__(self, asset_id: str):
        """Parser for asset id.

        :param asset_id: The asset id.
        :type asset_id: str
        :raises ~azure.ai.ml.exceptions.ValidationException~: Raised if the asset id is incorrectly formatted.
        """
        match = re.match(AMLAssetId.REGEX_PATTERN, asset_id)
        if match is None:
            msg = "Invalid AzureML Asset Id {}"
            raise ValidationException(
                message=msg.format(asset_id),
                no_personal_data_message=msg.format("[asset_id]"),
                error_type=ValidationErrorType.INVALID_VALUE,
                error_category=ErrorCategory.USER_ERROR,
                target=ErrorTarget.ASSET,
            )

        self.location = match.group(1)
        self.workspace_id = match.group(2)
        self.asset_type = match.group(3)
        self.asset_name = match.group(4)
        self.asset_version = match.group(5)


class AzureResourceId:
    """Parser for a non-AzureML ARM Id.

    :param arm_id: The non-AzureML ARM id.
    :type arm_id: str
    :raises ~azure.ai.ml.exceptions.ValidationException~: Raised if the ARM id is incorrectly formatted.
    """

    REGEX_PATTERN = "^/?subscriptions/([^/]+)/resourceGroups/([^/]+)/providers/Microsoft.([^/]+)/([^/]+)/([^/]+)"
    RESOURCEGROUP_PATTERN = "^/?subscriptions/([^/]+)/providers/Microsoft.([^/]+)/([^/]+)/([^/]+)"

    def __init__(self, arm_id=None):
        if arm_id:
            match = re.match(AzureResourceId.REGEX_PATTERN, arm_id)
            rg_match = re.match(AzureResourceId.RESOURCEGROUP_PATTERN, arm_id)
            if match:
                self.subscription_id = match.group(1)
                self.resource_group_name = match.group(2)
                self.provider_namespace_with_type = match.group(3) + match.group(4)
                self.asset_type = match.group(4)
                self.asset_name = match.group(5)
            elif rg_match:
                self.subscription_id = rg_match.group(1)
                self.resource_group_name = None
                self.asset_name = rg_match.group(4)
                self.asset_type = rg_match.group(3)
            else:
                msg = "Invalid ARM Id {}"
                raise ValidationException(
                    message=msg.format(arm_id),
                    no_personal_data_message=msg.format("[arm_id]"),
                    error_type=ValidationErrorType.INVALID_VALUE,
                    error_category=ErrorCategory.USER_ERROR,
                    target=ErrorTarget.ARM_RESOURCE,
                )


class AzureStorageContainerResourceId:
    """Parser for a Azure Storage Container ARM id.

    :param arm_id: The Azure Storage Container ARM id.
    :type arm_id: str
    :raises ~azure.ai.ml.exceptions.ValidationException~: Raised if the ARM id is incorrectly formatted.
    """

    REGEX_PATTERN = (
        "^/?subscriptions/([^/]+)/resourceGroups/([^/]+)/providers/Microsoft.Storage"
        "/storageAccounts/([^/]+)/blobServices/default/containers/([^/]+)"
    )

    def __init__(self, arm_id=None):
        if arm_id:
            match = re.match(AzureStorageContainerResourceId.REGEX_PATTERN, arm_id)
            if match:
                self.subscription_id = match.group(1)
                self.resource_group_name = match.group(2)
                self.storage_account = match.group(3)
                self.container = match.group(4)
            else:
                msg = "Invalid Azure Storage Container Resource Id {}"
                raise ValidationException(
                    message=msg.format(arm_id),
                    no_personal_data_message=msg.format("[arm_id]"),
                    error_type=ValidationErrorType.INVALID_VALUE,
                    error_category=ErrorCategory.USER_ERROR,
                    target=ErrorTarget.ARM_RESOURCE,
                )


def _parse_endpoint_name_from_deployment_id(deployment_id: str) -> str:
    REGEX_PATTERN = (
        "^/?subscriptions/([^/]+)/resourceGroups/(["
        "^/]+)/providers/Microsoft.MachineLearningServices/workspaces/([^/]+)/([^/]+)/([^/]+)/deployments/([^/]+)"
    )
    match = re.match(REGEX_PATTERN, deployment_id)
    if match is None:
        msg = "Invalid Deployment Id {}"
        raise ValidationException(
            message=msg.format(deployment_id),
            no_personal_data_message=msg.format("[id]"),
            error_type=ValidationErrorType.INVALID_VALUE,
            error_category=ErrorCategory.USER_ERROR,
            target=ErrorTarget.DEPLOYMENT,
        )
    return match.group(5)


def parse_AzureML_id(name: str) -> Tuple[str, str, str]:
    if name.startswith(ARM_ID_PREFIX):
        name = name[len(ARM_ID_PREFIX) :]

    at_splits = name.rsplit("@", 1)
    if len(at_splits) > 1:
        return at_splits[0], None, name[1]
    colon_splits = name.rsplit(":", 1)
    return (
        colon_splits[0],
        None if len(colon_splits) == 1 else colon_splits[1],
        None,
    )


def parse_prefixed_name_version(name: str) -> Tuple[str, Optional[str]]:
    if name.startswith(ARM_ID_PREFIX):
        return parse_name_version(name[len(ARM_ID_PREFIX) :])
    return parse_name_version(name)


def parse_name_version(name: str) -> Tuple[str, Optional[str]]:
    if name.find("/") != -1 and name[0] != "/":
        raise ValidationException(
            f"Could not parse {name}. If providing an ARM id, it should start with a '/'.",
            no_personal_data_message=f"Could not parse {name}.",
            error_type=ValidationErrorType.INVALID_VALUE,
            error_category=ErrorCategory.USER_ERROR,
            target=ErrorTarget.ARM_RESOURCE,
        )
    token_list = name.split(":")
    if len(token_list) == 1:
        return name, None
    name, *version = token_list  # type: ignore
    return name, ":".join(version)


def parse_name_label(name: Optional[str]) -> Tuple[str, Optional[str]]:
    if not name or (name.find("/") != -1 and name[0] != "/"):
        raise ValidationException(
            f"Could not parse {name}. If providing an ARM id, it should start with a '/'.",
            no_personal_data_message=f"Could not parse {name}.",
            error_type=ValidationErrorType.INVALID_VALUE,
            error_category=ErrorCategory.USER_ERROR,
            target=ErrorTarget.ARM_RESOURCE,
        )
    token_list = name.rpartition("@")
    if not token_list[1]:  # separator not found
        *_, name = token_list
        return name, None
    name, _, label = token_list
    return name, label


def is_ARM_id_for_resource(name: Any, resource_type: str = ".*", sub_workspace_resource: bool = True) -> bool:
    if sub_workspace_resource:
        resource_regex = NAMED_RESOURCE_ID_FORMAT.format(
            ".*", ".*", AZUREML_RESOURCE_PROVIDER, ".*", resource_type, ".*"
        )
    else:
        resource_regex = LEVEL_ONE_NAMED_RESOURCE_ID_FORMAT.format(
            ".*", ".*", AZUREML_RESOURCE_PROVIDER, resource_type, ".*"
        )
    if isinstance(name, str) and re.match(resource_regex, name, re.IGNORECASE):
        return True
    return False


def is_ARM_id_for_parented_resource(name: str, parent_resource_type: str, child_resource_type: str) -> bool:
    resource_regex = NAMED_RESOURCE_ID_FORMAT_WITH_PARENT.format(
        ".*",
        ".*",
        AZUREML_RESOURCE_PROVIDER,
        ".*",
        parent_resource_type,
        ".*",
        child_resource_type,
        "*",
    )
    return re.match(resource_regex, name, re.IGNORECASE) is not None


def is_registry_id_for_resource(name: Any) -> bool:
    if isinstance(name, str) and re.match(REGISTRY_URI_REGEX_FORMAT, name, re.IGNORECASE):
        return True
    return False


def is_singularity_id_for_resource(name: Any) -> bool:
    if isinstance(name, str) and re.match(SINGULARITY_ID_REGEX_FORMAT, name, re.IGNORECASE):
        return True
    return False


def is_singularity_full_name_for_resource(name: Any) -> bool:
    if isinstance(name, str) and (re.match(SINGULARITY_FULL_NAME_REGEX_FORMAT, name, re.IGNORECASE)):
        return True
    return False


def is_singularity_short_name_for_resource(name: Any) -> bool:
    if isinstance(name, str) and re.match(SINGULARITY_SHORT_NAME_REGEX_FORMAT, name, re.IGNORECASE):
        return True
    return False


def get_arm_id_with_version(
    operation_scope: OperationScope,
    provider_name: str,
    provider_value: str,
    provider_version: str,
):
    return PROVIDER_RESOURCE_ID_WITH_VERSION.format(
        operation_scope.subscription_id,
        operation_scope.resource_group_name,
        operation_scope.workspace_name,
        provider_name,
        provider_value,
        provider_version,
    )


def generate_data_arm_id(operation_scope: OperationScope, name: str, version: int):
    return get_arm_id_with_version(operation_scope, DATA_ARM_TYPE, name, str(version))


def remove_aml_prefix(resource_id: Optional[str]) -> Optional[str]:
    if not resource_id:
        return None

    if resource_id.startswith(ARM_ID_PREFIX):
        return resource_id[len(ARM_ID_PREFIX) :]
    return resource_id


def get_resource_name_from_arm_id(resource_id: str) -> str:
    # parse arm id to datastore name
    return AMLNamedArmId(resource_id).asset_name


def get_resource_name_from_arm_id_safe(resource_id: str) -> str:
    """Get the resource name from an ARM id.

    :param resource_id: The resource's ARM ID
    :type resource_id: str
    :return:
      * Resource Name if input string is ARM id
      * Original input otherwise
    :rtype: str
    """
    try:
        return get_resource_name_from_arm_id(resource_id)
    except ValidationException:
        # already a name
        return resource_id
    except AttributeError:
        # None or empty string
        return resource_id
    except Exception:  # pylint: disable=W0718
        # unexpected error
        module_logger.warning("Failed to parse resource id: %s", resource_id)
        return resource_id


def get_arm_id_object_from_id(
    resource_id: str,
) -> Union[AMLVersionedArmId, AMLNamedArmId, AzureResourceId]:
    """Attempts to create and return one of: AMLVersionedId, AMLNamedId, AzureResoureId. In the case than an AzureML ARM
    Id is passed in, either AMLVersionedId or AMLNamedId will be created depending on resource type In the case that a
    non-AzureML ARM id is passed in, an AzureResourceId will be returned.

    :param resource_id: the ARM Id to parse
    :type resource_id: str
    :raises ~azure.ai.ml.exceptions.ValidationException~: Raised if the ARM id is incorrectly formatted.
    :return: The parser for the given ARM Id
    :rtype: Union[AMLVersionedArmId, AMLNamedArmId, AzureResourceId]
    """

    # Try each type of parser for the ARM id. If none succeed, raise a ValueError
    try:
        return AMLVersionedArmId(resource_id)
    except ValidationException:
        pass

    try:
        return AMLNamedArmId(resource_id)
    except ValidationException:
        pass

    try:
        return AzureResourceId(resource_id)
    except ValidationException:
        pass

    msg = "Invalid ARM Id {}"
    raise ValidationException(
        message=msg.format(resource_id),
        no_personal_data_message=msg.format("[resource_id]"),
        error_type=ValidationErrorType.INVALID_VALUE,
        error_category=ErrorCategory.USER_ERROR,
        target=ErrorTarget.DEPLOYMENT,
    )


def remove_datastore_prefix(datastore_id: Optional[str]) -> Optional[str]:
    if not datastore_id:
        return None

    if datastore_id.startswith(DATASTORE_SHORT_URI):
        return datastore_id[len(DATASTORE_SHORT_URI) :]
    return datastore_id



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=unused-argument

from marshmallow import fields, post_load

from azure.ai.ml._schema.core.schema import PatchedSchemaMeta


class AlertNotificationSchema(metaclass=PatchedSchemaMeta):
    emails = fields.List(fields.Str)

    @post_load
    def make(self, data, **kwargs):
        from azure.ai.ml.entities._monitoring.alert_notification import AlertNotification

        return AlertNotification(**data)



# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------

# pylint: disable=protected-access,broad-except

import random
import re
import subprocess
from typing import Any, Dict, Optional

from marshmallow.exceptions import ValidationError as SchemaValidationError

from azure.ai.ml._exception_helper import log_and_raise_error
from azure.ai.ml._local_endpoints import LocalEndpointMode
from azure.ai.ml._restclient.v2022_02_01_preview.models import DeploymentLogsRequest
from azure.ai.ml._restclient.v2023_04_01_preview import AzureMachineLearningWorkspaces as ServiceClient042023Preview
from azure.ai.ml._scope_dependent_operations import (
    OperationConfig,
    OperationsContainer,
    OperationScope,
    _ScopeDependentOperations,
)
from azure.ai.ml._telemetry import ActivityType, monitor_with_activity
from azure.ai.ml._utils._arm_id_utils import AMLVersionedArmId
from azure.ai.ml._utils._azureml_polling import AzureMLPolling
from azure.ai.ml._utils._endpoint_utils import upload_dependencies, validate_scoring_script
from azure.ai.ml._utils._logger_utils import OpsLogger
from azure.ai.ml._utils._package_utils import package_deployment
from azure.ai.ml.constants._common import ARM_ID_PREFIX, AzureMLResourceType, LROConfigurations
from azure.ai.ml.constants._deployment import DEFAULT_MDC_PATH, EndpointDeploymentLogContainerType, SmallSKUs
from azure.ai.ml.entities import Data, OnlineDeployment
from azure.ai.ml.exceptions import (
    ErrorCategory,
    ErrorTarget,
    InvalidVSCodeRequestError,
    LocalDeploymentGPUNotAvailable,
    ValidationErrorType,
    ValidationException,
)
from azure.core.credentials import TokenCredential
from azure.core.paging import ItemPaged
from azure.core.polling import LROPoller
from azure.core.tracing.decorator import distributed_trace

from ._local_deployment_helper import _LocalDeploymentHelper
from ._operation_orchestrator import OperationOrchestrator

ops_logger = OpsLogger(__name__)
module_logger = ops_logger.module_logger


class OnlineDeploymentOperations(_ScopeDependentOperations):
    """OnlineDeploymentOperations.

    You should not instantiate this class directly. Instead, you should create an MLClient instance that instantiates it
    for you and attaches it as an attribute.
    """

    def __init__(
        self,
        operation_scope: OperationScope,
        operation_config: OperationConfig,
        service_client_04_2023_preview: ServiceClient042023Preview,
        all_operations: OperationsContainer,
        local_deployment_helper: _LocalDeploymentHelper,
        credentials: Optional[TokenCredential] = None,
        **kwargs: Dict,
    ):
        super(OnlineDeploymentOperations, self).__init__(operation_scope, operation_config)
        ops_logger.update_filter()
        self._local_deployment_helper = local_deployment_helper
        self._online_deployment = service_client_04_2023_preview.online_deployments
        self._online_endpoint_operations = service_client_04_2023_preview.online_endpoints
        self._all_operations = all_operations
        self._credentials = credentials
        self._init_kwargs = kwargs

    @distributed_trace
    @monitor_with_activity(ops_logger, "OnlineDeployment.BeginCreateOrUpdate", ActivityType.PUBLICAPI)
    def begin_create_or_update(
        self,
        deployment: OnlineDeployment,
        *,
        local: bool = False,
        vscode_debug: bool = False,
        skip_script_validation: bool = False,
        local_enable_gpu: bool = False,
        **kwargs: Any,
    ) -> LROPoller[OnlineDeployment]:
        """Create or update a deployment.

        :param deployment: the deployment entity
        :type deployment: ~azure.ai.ml.entities.OnlineDeployment
        :keyword local: Whether deployment should be created locally, defaults to False
        :paramtype local: bool
        :keyword vscode_debug: Whether to open VSCode instance to debug local deployment, defaults to False
        :paramtype vscode_debug: bool
        :keyword skip_script_validation: Whether or not to skip validation of the deployment script. Defaults to False.
        :paramtype skip_script_validation: bool
        :keyword local_enable_gpu: enable local container to access gpu
        :paramtype local_enable_gpu: bool
        :raises ~azure.ai.ml.exceptions.ValidationException: Raised if OnlineDeployment cannot
            be successfully validated. Details will be provided in the error message.
        :raises ~azure.ai.ml.exceptions.AssetException: Raised if OnlineDeployment assets
            (e.g. Data, Code, Model, Environment) cannot be successfully validated.
            Details will be provided in the error message.
        :raises ~azure.ai.ml.exceptions.ModelException: Raised if OnlineDeployment model cannot be
            successfully validated. Details will be provided in the error message.
        :raises ~azure.ai.ml.exceptions.DeploymentException: Raised if OnlineDeployment type is unsupported.
            Details will be provided in the error message.
        :raises ~azure.ai.ml.exceptions.LocalEndpointNotFoundError: Raised if local endpoint resource does not exist.
        :raises ~azure.ai.ml.exceptions.LocalEndpointInFailedStateError: Raised if local endpoint is in a failed state.
        :raises ~azure.ai.ml.exceptions.InvalidLocalEndpointError: Raised if Docker image cannot be
            found for local deployment.
        :raises ~azure.ai.ml.exceptions.LocalEndpointImageBuildError: Raised if Docker image cannot be
            successfully built for local deployment.
        :raises ~azure.ai.ml.exceptions.RequiredLocalArtifactsNotFoundError: Raised if local artifacts cannot be
            found for local deployment.
        :raises ~azure.ai.ml.exceptions.InvalidVSCodeRequestError: Raised if VS Debug is invoked with a remote endpoint.
            VSCode debug is only supported for local endpoints.
        :raises ~azure.ai.ml.exceptions.LocalDeploymentGPUNotAvailable: Raised if Nvidia GPU is not available in the
            system and local_enable_gpu is set while local deployment
        :raises ~azure.ai.ml.exceptions.VSCodeCommandNotFound: Raised if VSCode instance cannot be instantiated.
        :return: A poller to track the operation status
        :rtype: ~azure.core.polling.LROPoller[~azure.ai.ml.entities.OnlineDeployment]
        """
        try:
            if vscode_debug and not local:
                raise InvalidVSCodeRequestError(
                    msg="VSCode Debug is only support for local endpoints. Please set local to True."
                )
            if local:
                if local_enable_gpu:
                    try:
                        subprocess.run("nvidia-smi", check=True)
                    except Exception as ex:
                        raise LocalDeploymentGPUNotAvailable(
                            msg=(
                                "Nvidia GPU is not available in your local system."
                                " Use nvidia-smi command to see the available GPU"
                            )
                        ) from ex
                return self._local_deployment_helper.create_or_update(
                    deployment=deployment,
                    local_endpoint_mode=self._get_local_endpoint_mode(vscode_debug),
                    local_enable_gpu=local_enable_gpu,
                )
            if deployment and deployment.instance_type and deployment.instance_type.lower() in SmallSKUs:
                module_logger.warning(
                    "Instance type %s may be too small for compute resources. "
                    "Minimum recommended compute SKU is Standard_DS3_v2 for general purpose endpoints. Learn more about SKUs here: "  # pylint: disable=line-too-long
                    "https://learn.microsoft.com/azure/machine-learning/referencemanaged-online-endpoints-vm-sku-list",
                    deployment.instance_type,
                )
            if (
                not skip_script_validation
                and deployment
                and deployment.code_configuration
                and not deployment.code_configuration.code.startswith(ARM_ID_PREFIX)  # type: ignore[union-attr]
                and not re.match(AMLVersionedArmId.REGEX_PATTERN, deployment.code_configuration.code)  # type: ignore
            ):
                validate_scoring_script(deployment)

            path_format_arguments = {
                "endpointName": deployment.name,
                "resourceGroupName": self._resource_group_name,
                "workspaceName": self._workspace_name,
            }

            # This get() is to ensure, the endpoint exists and fail before even start the deployment
            module_logger.info("Check: endpoint %s exists", deployment.endpoint_name)
            self._online_endpoint_operations.get(
                resource_group_name=self._resource_group_name,
                workspace_name=self._workspace_name,
                endpoint_name=deployment.endpoint_name,
            )
            orchestrators = OperationOrchestrator(
                operation_container=self._all_operations,
                operation_scope=self._operation_scope,
                operation_config=self._operation_config,
            )
            if deployment.data_collector:
                self._register_collection_data_assets(deployment=deployment)

            upload_dependencies(deployment, orchestrators)
            try:
                location = self._get_workspace_location()
                is_package_model = deployment.package_model if hasattr(deployment, "package_model") else False
                if kwargs.pop("package_model", False) or is_package_model:
                    deployment = package_deployment(deployment, self._all_operations.all_operations["models"])
                    module_logger.info("\nStarting deployment")

                deployment_rest = deployment._to_rest_object(location=location)  # type: ignore

                poller = self._online_deployment.begin_create_or_update(
                    resource_group_name=self._resource_group_name,
                    workspace_name=self._workspace_name,
                    endpoint_name=deployment.endpoint_name,
                    deployment_name=deployment.name,
                    body=deployment_rest,
                    polling=AzureMLPolling(
                        LROConfigurations.POLL_INTERVAL,
                        path_format_arguments=path_format_arguments,
                        **self._init_kwargs,
                    ),
                    polling_interval=LROConfigurations.POLL_INTERVAL,
                    **self._init_kwargs,
                    cls=lambda response, deserialized, headers: OnlineDeployment._from_rest_object(deserialized),
                )
                return poller
            except Exception as ex:
                raise ex
        except Exception as ex:  # pylint: disable=W0718
            if isinstance(ex, (ValidationException, SchemaValidationError)):
                log_and_raise_error(ex)
            else:
                raise ex

    @distributed_trace
    @monitor_with_activity(ops_logger, "OnlineDeployment.Get", ActivityType.PUBLICAPI)
    def get(self, name: str, endpoint_name: str, *, local: Optional[bool] = False) -> OnlineDeployment:
        """Get a deployment resource.

        :param name: The name of the deployment
        :type name: str
        :param endpoint_name: The name of the endpoint
        :type endpoint_name: str
        :keyword local: Whether deployment should be retrieved from local docker environment, defaults to False
        :paramtype local: Optional[bool]
        :raises ~azure.ai.ml.exceptions.LocalEndpointNotFoundError: Raised if local endpoint resource does not exist.
        :return: a deployment entity
        :rtype: ~azure.ai.ml.entities.OnlineDeployment
        """
        if local:
            deployment = self._local_deployment_helper.get(endpoint_name=endpoint_name, deployment_name=name)
        else:
            deployment = OnlineDeployment._from_rest_object(
                self._online_deployment.get(
                    endpoint_name=endpoint_name,
                    deployment_name=name,
                    resource_group_name=self._resource_group_name,
                    workspace_name=self._workspace_name,
                    **self._init_kwargs,
                )
            )

        deployment.endpoint_name = endpoint_name
        return deployment

    @distributed_trace
    @monitor_with_activity(ops_logger, "OnlineDeployment.Delete", ActivityType.PUBLICAPI)
    def begin_delete(self, name: str, endpoint_name: str, *, local: Optional[bool] = False) -> LROPoller[None]:
        """Delete a deployment.

        :param name: The name of the deployment
        :type name: str
        :param endpoint_name: The name of the endpoint
        :type endpoint_name: str
        :keyword local: Whether deployment should be retrieved from local docker environment, defaults to False
        :paramtype local: Optional[bool]
        :raises ~azure.ai.ml.exceptions.LocalEndpointNotFoundError: Raised if local endpoint resource does not exist.
        :return: A poller to track the operation status
        :rtype: ~azure.core.polling.LROPoller[None]
        """
        if local:
            return self._local_deployment_helper.delete(name=endpoint_name, deployment_name=name)
        return self._online_deployment.begin_delete(
            endpoint_name=endpoint_name,
            deployment_name=name,
            resource_group_name=self._resource_group_name,
            workspace_name=self._workspace_name,
            **self._init_kwargs,
        )

    @distributed_trace
    @monitor_with_activity(ops_logger, "OnlineDeployment.GetLogs", ActivityType.PUBLICAPI)
    def get_logs(
        self,
        name: str,
        endpoint_name: str,
        lines: int,
        *,
        container_type: Optional[str] = None,
        local: bool = False,
    ) -> str:
        """Retrive the logs from online deployment.

        :param name: The name of the deployment
        :type name: str
        :param endpoint_name: The name of the endpoint
        :type endpoint_name: str
        :param lines: The maximum number of lines to tail
        :type lines: int
        :keyword container_type: The type of container to retrieve logs from. Possible values include:
            "StorageInitializer", "InferenceServer", defaults to None
        :type container_type: Optional[str]
        :keyword local: [description], defaults to False
        :paramtype local: bool
        :return: the logs
        :rtype: str
        """
        if local:
            return self._local_deployment_helper.get_deployment_logs(
                endpoint_name=endpoint_name, deployment_name=name, lines=lines
            )
        if container_type:
            container_type = self._validate_deployment_log_container_type(container_type)  # type: ignore
        log_request = DeploymentLogsRequest(container_type=container_type, tail=lines)
        return str(
            self._online_deployment.get_logs(
                resource_group_name=self._resource_group_name,
                workspace_name=self._workspace_name,
                endpoint_name=endpoint_name,
                deployment_name=name,
                body=log_request,
                **self._init_kwargs,
            ).content
        )

    @distributed_trace
    @monitor_with_activity(ops_logger, "OnlineDeployment.List", ActivityType.PUBLICAPI)
    def list(self, endpoint_name: str, *, local: bool = False) -> ItemPaged[OnlineDeployment]:
        """List a deployment resource.

        :param endpoint_name: The name of the endpoint
        :type endpoint_name: str
        :keyword local: Whether deployment should be retrieved from local docker environment, defaults to False
        :paramtype local: bool
        :return: an iterator of deployment entities
        :rtype: Iterable[~azure.ai.ml.entities.OnlineDeployment]
        """
        if local:
            return self._local_deployment_helper.list()
        return self._online_deployment.list(
            endpoint_name=endpoint_name,
            resource_group_name=self._resource_group_name,
            workspace_name=self._workspace_name,
            cls=lambda objs: [OnlineDeployment._from_rest_object(obj) for obj in objs],
            **self._init_kwargs,
        )

    def _validate_deployment_log_container_type(self, container_type: EndpointDeploymentLogContainerType) -> str:
        if container_type == EndpointDeploymentLogContainerType.INFERENCE_SERVER:
            return EndpointDeploymentLogContainerType.INFERENCE_SERVER_REST

        if container_type == EndpointDeploymentLogContainerType.STORAGE_INITIALIZER:
            return EndpointDeploymentLogContainerType.STORAGE_INITIALIZER_REST

        msg = "Invalid container type '{}'. Supported container types are {} and {}"
        msg = msg.format(
            container_type,
            EndpointDeploymentLogContainerType.INFERENCE_SERVER,
            EndpointDeploymentLogContainerType.STORAGE_INITIALIZER,
        )
        raise ValidationException(
            message=msg,
            target=ErrorTarget.ONLINE_DEPLOYMENT,
            no_personal_data_message=msg,
            error_category=ErrorCategory.USER_ERROR,
            error_type=ValidationErrorType.INVALID_VALUE,
        )

    def _get_ARM_deployment_name(self, name: str) -> str:
        random.seed(version=2)
        return f"{self._workspace_name}-{name}-{random.randint(1, 10000000)}"

    def _get_workspace_location(self) -> str:
        """Get the workspace location

        TODO[TASK 1260265]: can we cache this information and only refresh when the operation_scope is changed?

        :return: The workspace location
        :rtype: str
        """
        return str(
            self._all_operations.all_operations[AzureMLResourceType.WORKSPACE].get(self._workspace_name).location
        )

    def _get_local_endpoint_mode(self, vscode_debug: Any) -> LocalEndpointMode:
        return LocalEndpointMode.VSCodeDevContainer if vscode_debug else LocalEndpointMode.DetachedContainer

    def _register_collection_data_assets(self, deployment: OnlineDeployment) -> None:
        for name, value in deployment.data_collector.collections.items():
            data_name = f"{deployment.endpoint_name}-{deployment.name}-{name}"
            data_version = "1"
            data_path = f"{DEFAULT_MDC_PATH}/{deployment.endpoint_name}/{deployment.name}/{name}"
            if value.data:
                if value.data.name:
                    data_name = value.data.name

                if value.data.version:
                    data_version = value.data.version

                if value.data.path:
                    data_path = value.data.path

            data_object = Data(
                name=data_name,
                version=data_version,
                path=data_path,
            )

            try:
                result = self._all_operations._all_operations[AzureMLResourceType.DATA].create_or_update(data_object)
            except Exception as e:
                if "already exists" in str(e):
                    result = self._all_operations._all_operations[AzureMLResourceType.DATA].get(data_name, data_version)
                else:
                    raise e
            deployment.data_collector.collections[name].data = (
                f"/subscriptions/{self._subscription_id}/resourceGroups/{self._resource_group_name}"
                f"/providers/Microsoft.MachineLearningServices/workspaces/{self._workspace_name}"
                f"/data/{result.name}/versions/{result.version}"
            )



from typing import Callable

import pytest
from devtools_testutils import AzureRecordedTestCase

from azure.ai.ml import MLClient
from azure.ai.ml.constants._common import LROConfigurations
from azure.ai.ml.entities._load_functions import load_schedule


@pytest.mark.timeout(1200)
@pytest.mark.e2etest
@pytest.mark.usefixtures("recorded_test", "mock_code_hash", "mock_asset_name", "mock_component_hash")
@pytest.mark.pipeline_test
class TestSchedule(AzureRecordedTestCase):
    @pytest.mark.parametrize(
        "source_type",
        [
            ("database"),
            ("file_system"),
        ],
    )
    def test_create_schedule_data_import(self, client: MLClient, randstr: Callable[[], str], source_type: str):
        params_override = [{"name": randstr("name")}]
        test_path = "./tests/test_configs/data_import/schedule_data_import_" + source_type + ".yml"
        schedule = load_schedule(test_path, params_override=params_override)
        rest_schedule = client.schedules.begin_create_or_update(schedule).result(
            timeout=LROConfigurations.POLLING_TIMEOUT
        )
        rest_schedule_dict = rest_schedule._to_dict()
        assert rest_schedule.name == schedule.name
        assert rest_schedule_dict["trigger"]["type"] == "cron"
        assert rest_schedule_dict["import_data"]["source"]["type"] == source_type
        assert rest_schedule_dict["import_data"]["path"] == "azureml://datastores/workspaceblobstore/paths/{name}"

        if source_type == "file_system":
            assert rest_schedule_dict["import_data"]["name"] == "my_s3_asset"
            assert rest_schedule_dict["import_data"]["type"] == "uri_folder"
            assert rest_schedule_dict["import_data"]["source"]["connection"] == "azureml:my_s3_connection"
            assert rest_schedule_dict["import_data"]["source"]["path"] == "test1/*"
        else:
            assert rest_schedule_dict["import_data"]["name"] == "my_azuresqldb_asset"
            assert rest_schedule_dict["import_data"]["type"] == "mltable"
            assert rest_schedule_dict["import_data"]["source"]["connection"] == "azureml:my_azuresqldb_connection"
            assert rest_schedule_dict["import_data"]["source"]["query"] == "select * from region"
